import asyncio
from decimal import Decimal
from typing import Dict, List, Literal
from nexustrader.constants import AccountType, SubmitType
from nexustrader.schema import OrderSubmit, InstrumentId
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.okx import OkxAccountType
from nexustrader.exchange.okx.schema import OkxMarket
from nexustrader.base import ExecutionManagementSystem
from nexustrader.schema import CancelAllOrderSubmit, CancelOrderSubmit


class OkxExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, OkxMarket]

    OKX_ACCOUNT_TYPE_PRIORITY = [
        OkxAccountType.DEMO,
        # OkxAccountType.AWS,
        OkxAccountType.LIVE,
    ]

    def __init__(
        self,
        market: Dict[str, OkxMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        registry: OrderRegistry,
        is_mock: bool = False,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            clock=clock,
            task_manager=task_manager,
            registry=registry,
            is_mock=is_mock,
        )
        self._okx_account_type: OkxAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, OkxAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()
                break

    def _set_account_type(self):
        account_types = self._private_connectors.keys()
        for account_type in self.OKX_ACCOUNT_TYPE_PRIORITY:
            if account_type in account_types:
                self._okx_account_type = account_type
                break

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        if self._is_mock:
            if instrument_id.is_spot:
                return OkxAccountType.SPOT_MOCK
            elif instrument_id.is_linear:
                return OkxAccountType.LINEAR_MOCK
            elif instrument_id.is_inverse:
                return OkxAccountType.INVERSE_MOCK
        else:
            return self._okx_account_type

    def _submit_order(
        self,
        order: OrderSubmit | List[OrderSubmit],
        submit_type: SubmitType,
        account_type: AccountType | None = None,
    ):
        if isinstance(order, list):
            if not account_type:
                account_type = self._instrument_id_to_account_type(
                    order[0].instrument_id
                )

            # Split batch orders into chunks of 20
            for i in range(0, len(order), 20):
                batch = order[i : i + 20]
                self._order_submit_queues[account_type].put_nowait((batch, submit_type))
        else:
            if not account_type:
                account_type = self._instrument_id_to_account_type(order.instrument_id)
            self._order_submit_queues[account_type].put_nowait((order, submit_type))

    def _get_min_order_amount(
        self, symbol: str, market: OkxMarket, px: float
    ) -> Decimal:
        min_order_amount = market.limits.amount.min
        min_order_amount = super()._amount_to_precision(
            symbol, min_order_amount, mode="ceil"
        )

        if not market.spot:
            # for linear and inverse, the min order amount is contract size and ctVal is base amount per contract
            min_order_amount *= Decimal(market.info.ctVal)

        return min_order_amount

    # override the base method
    def _amount_to_precision(
        self,
        symbol: str,
        amount: float,
        mode: Literal["round"] | Literal["ceil"] | Literal["floor"] = "round",
    ) -> Decimal:
        market = self._market[symbol]
        ctVal = Decimal("1")
        if not market.spot:
            ctVal = Decimal(market.info.ctVal)
            amount = Decimal(str(amount)) / ctVal
        return super()._amount_to_precision(symbol, amount, mode) * ctVal

    async def _cancel_all_orders(
        self, order_submit: CancelAllOrderSubmit, account_type: AccountType
    ):
        # override the base methods
        symbol = order_submit.symbol
        oids = self._cache.get_open_orders(symbol)
        for oid in oids:
            order_submit = CancelOrderSubmit(
                symbol=symbol,
                instrument_id=InstrumentId.from_str(symbol),
                oid=oid,
            )
            await self._cancel_order(order_submit, account_type)



from typing import Callable, List, Dict
from typing import Any
from urllib.parse import urlencode
from nexustrader.base import WSClient
from nexustrader.exchange.binance.constants import (
    BinanceAccountType,
    BinanceKlineInterval,
    BinanceRateLimiter,
)
from nexustrader.core.entity import TaskManager
from nexustrader.core.nautilius_core import hmac_signature, LiveClock


class BinanceWSClient(WSClient):
    def __init__(
        self,
        account_type: BinanceAccountType,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        ws_suffix: str = "/ws",
        custom_url: str | None = None,
    ):
        self._account_type = account_type
        url = account_type.ws_url

        if ws_suffix not in ["/ws", "/stream"]:
            raise ValueError(f"Invalid ws_suffix: {ws_suffix}")

        url += ws_suffix

        if custom_url is not None:
            url = custom_url

        super().__init__(
            url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            enable_auto_ping=False,
        )

    def _send_payload(
        self, params: List[str], method: str = "SUBSCRIBE", chunk_size: int = 50
    ):
        # Split params into chunks of 100 if length exceeds 100
        params_chunks = [
            params[i : i + chunk_size] for i in range(0, len(params), chunk_size)
        ]

        for chunk in params_chunks:
            payload = {
                "method": method,
                "params": chunk,
                "id": self._clock.timestamp_ms(),
            }
            self._send(payload)

    async def _subscribe(self, params: List[str]):
        params = [param for param in params if param not in self._subscriptions]

        if not params:
            return

        for param in params:
            self._subscriptions.append(param)
            self._log.debug(f"Subscribing to {param}...")

        await self.connect()
        self._send_payload(params, method="SUBSCRIBE")

    async def _unsubscribe(self, params: List[str]):
        params = [param for param in params if param in self._subscriptions]

        if not params:
            return

        for param in params:
            self._subscriptions.remove(param)
            self._log.debug(f"Unsubscribing from {param}...")

        await self.connect()
        self._send_payload(params, method="UNSUBSCRIBE")

    async def subscribe_agg_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@aggTrade" for symbol in symbols]
        await self._subscribe(params)

    async def subscribe_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@trade" for symbol in symbols]
        await self._subscribe(params)

    async def subscribe_book_ticker(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@bookTicker" for symbol in symbols]
        await self._subscribe(params)

    async def subscribe_partial_book_depth(self, symbols: List[str], level: int):
        if level not in (5, 10, 20):
            raise ValueError("Level must be 5, 10, or 20")
        params = [f"{symbol.lower()}@depth{level}@100ms" for symbol in symbols]
        await self._subscribe(params)

    async def subscribe_mark_price(self, symbols: List[str]):
        if not self._account_type.is_future:
            raise ValueError("Only Supported for `Future Account`")
        params = [f"{symbol.lower()}@markPrice@1s" for symbol in symbols]
        await self._subscribe(params)

    async def subscribe_user_data_stream(self, listen_key: str):
        await self._subscribe([listen_key])

    async def subscribe_kline(
        self,
        symbols: List[str],
        interval: BinanceKlineInterval,
    ):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@kline_{interval.value}" for symbol in symbols]
        await self._subscribe(params)

    async def unsubscribe_agg_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@aggTrade" for symbol in symbols]
        await self._unsubscribe(params)

    async def unsubscribe_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@trade" for symbol in symbols]
        await self._unsubscribe(params)

    async def unsubscribe_book_ticker(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@bookTicker" for symbol in symbols]
        await self._unsubscribe(params)

    async def unsubscribe_partial_book_depth(self, symbols: List[str], level: int):
        if level not in (5, 10, 20):
            raise ValueError("Level must be 5, 10, or 20")
        params = [f"{symbol.lower()}@depth{level}@100ms" for symbol in symbols]
        await self._unsubscribe(params)

    async def unsubscribe_mark_price(self, symbols: List[str]):
        if not self._account_type.is_future:
            raise ValueError("Only Supported for `Future Account`")
        params = [f"{symbol.lower()}@markPrice@1s" for symbol in symbols]
        await self._unsubscribe(params)

    async def unsubscribe_user_data_stream(self, listen_key: str):
        await self._unsubscribe([listen_key])

    async def unsubscribe_kline(
        self,
        symbols: List[str],
        interval: BinanceKlineInterval,
    ):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@kline_{interval.value}" for symbol in symbols]
        await self._unsubscribe(params)

    async def _resubscribe(self):
        self._send_payload(self._subscriptions)


class BinanceWSApiClient(WSClient):
    def __init__(
        self,
        account_type: BinanceAccountType,
        api_key: str,
        secret: str,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        enable_rate_limit: bool,
    ):
        self._account_type = account_type
        self._api_key = api_key
        self._secret = secret
        self._limiter = BinanceRateLimiter(enable_rate_limit)

        url = account_type.ws_order_url

        if not url:
            raise ValueError(f"WebSocket URL not supported for {account_type}")

        super().__init__(
            url=url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            enable_auto_ping=False,
        )

    def _generate_signature_v2(self, query: str) -> str:
        signature = hmac_signature(self._secret, query)
        return signature

    def _send_payload(
        self,
        id: str,
        method: str,
        params: Dict[str, Any],
        required_ts: bool = True,
        auth: bool = True,
    ):
        if required_ts:
            params["timestamp"] = self._clock.timestamp_ms()

        if auth:
            params["apiKey"] = self._api_key
            query = urlencode(sorted(params.items()))
            signature = self._generate_signature_v2(query)
            params["signature"] = signature

        payload = {
            "method": method,
            "id": id,
            "params": params,
        }
        self._send(payload)

    async def spot_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.api_order_limit(cost=1)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def spot_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.api_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    async def usdm_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.fapi_order_limit(cost=0)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def usdm_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.fapi_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    async def coinm_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.dapi_order_limit(cost=0)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def coinm_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.dapi_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    async def _resubscribe(self):
        pass


import asyncio  # noqa


async def main():
    from nexustrader.constants import settings
    from nexustrader.core.entity import TaskManager
    from nexustrader.core.nautilius_core import LiveClock, setup_nautilus_core, UUID4

    # API_KEY = settings.BINANCE.SPOT.TESTNET.API_KEY
    # SECRET = settings.BINANCE.SPOT.TESTNET.SECRET

    API_KEY = settings.BINANCE.FUTURE.TESTNET_1.API_KEY
    SECRET = settings.BINANCE.FUTURE.TESTNET_1.SECRET

    log_guard = setup_nautilus_core(  # noqa
        trader_id="bnc-test",
        level_stdout="DEBUG",
    )

    task_manager = TaskManager(
        loop=asyncio.get_event_loop(),
    )

    ws_api_client = BinanceWSApiClient(
        account_type=BinanceAccountType.USD_M_FUTURE_TESTNET,
        api_key=API_KEY,
        secret=SECRET,
        handler=lambda msg: print(msg),
        task_manager=task_manager,
        clock=LiveClock(),
        enable_rate_limit=True,
    )

    await ws_api_client.connect()

    await ws_api_client.usdm_new_order(
        id=UUID4().value,
        symbol="BTCUSDT",
        side="BUY",
        type="LIMIT",
        quantity="0.003",
        price="120000",
        # timeInForce="GTC",
    )

    # await ws_api_client.usdm_cancel_order(
    #     id="aa510a1f-7240-4368-8cc0-ba577483a734",
    #     symbol="BTCUSDT",
    #     orderId=5594834544,  # Replace with a valid order ID
    # )

    # await ws_api_client.spot_new_order(
    #     id=UUID4().value,
    #     symbol="BTCUSDT",
    #     side="BUY",
    #     type="LIMIT",
    #     quantity="0.001",
    #     price="80000",
    #     timeInForce="GTC",
    # )
    # await ws_api_client.spot_cancel_open_orders(
    #     id="20dbbfdb-fc47-4abf-bb0f-bc0cd117c29a",
    #     symbol="BTCUSDT",
    # )

    await task_manager.wait()


if __name__ == "__main__":
    asyncio.run(main())



import aiosqlite
import sqlite3
from pathlib import Path
from typing import Dict, Set, List, Optional, Any

from nexustrader.backends.db import StorageBackend
from nexustrader.schema import Order, Position, AlgoOrder, Balance, AccountBalance
from nexustrader.constants import AccountType, ExchangeType


class SQLiteBackend(StorageBackend):
    def __init__(
        self,
        strategy_id: str,
        user_id: str,
        table_prefix: str,
        log,
        db_path: str = ".keys/cache.db",
        **kwargs,
    ):
        super().__init__(strategy_id, user_id, table_prefix, log, **kwargs)
        self.db_path = db_path
        self._db_async = None
        self._db = None

    async def _init_conn(self) -> None:
        db_path = Path(self.db_path)
        db_path.parent.mkdir(parents=True, exist_ok=True)
        self._db_async = await aiosqlite.connect(str(db_path))
        self._db = sqlite3.connect(str(db_path))

    async def _init_table(self) -> None:
        async with self._db_async.cursor() as cursor:
            await cursor.executescript(f"""
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_orders (
                    timestamp INTEGER,
                    oid TEXT PRIMARY KEY,
                    eid TEXT,
                    symbol TEXT,
                    side TEXT, 
                    type TEXT,
                    amount TEXT,
                    price REAL,
                    status TEXT,
                    fee TEXT,
                    fee_currency TEXT,
                    data BLOB
                );
                
                CREATE INDEX IF NOT EXISTS idx_orders_symbol 
                ON {self.table_prefix}_orders(symbol);
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_algo_orders (
                    timestamp INTEGER,
                    oid TEXT PRIMARY KEY,
                    symbol TEXT,
                    data BLOB
                );
                
                CREATE INDEX IF NOT EXISTS idx_algo_orders_symbol 
                ON {self.table_prefix}_algo_orders(symbol);
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_positions (
                    symbol PRIMARY KEY,
                    exchange TEXT,
                    side TEXT,
                    amount TEXT,
                    data BLOB
                );
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_open_orders (
                    oid PRIMARY KEY,
                    eid TEXT,
                    exchange TEXT,
                    symbol TEXT
                );
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_balances (
                    asset TEXT,
                    account_type TEXT,
                    free TEXT,
                    locked TEXT,
                    PRIMARY KEY (asset, account_type)
                );
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_pnl (
                    timestamp INTEGER PRIMARY KEY,
                    pnl REAL,
                    unrealized_pnl REAL
                );
                
                CREATE TABLE IF NOT EXISTS {self.table_prefix}_params (
                    key TEXT PRIMARY KEY,
                    value BLOB,
                    timestamp INTEGER DEFAULT (strftime('%s', 'now') * 1000)
                );
            """)
            await self._db_async.commit()

    def _get_cursor(self):
        """Get a thread-safe SQLite cursor, creating new connection if needed."""
        try:
            return self._db.cursor()
        except sqlite3.ProgrammingError:
            # SQLite connection is from different thread, create new one
            db_path = Path(self.db_path)
            return sqlite3.connect(str(db_path)).cursor()

    async def close(self) -> None:
        if self._db_async:
            await self._db_async.close()
        if self._db:
            self._db.close()

    async def sync_orders(self, mem_orders: Dict[str, Order]) -> None:
        async with self._db_async.cursor() as cursor:
            for oid, order in mem_orders.copy().items():
                await cursor.execute(
                    f"INSERT OR REPLACE INTO {self.table_prefix}_orders "
                    "(timestamp, oid, eid, symbol, side, type, amount, price, status, fee, fee_currency, data) "
                    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        order.timestamp,
                        order.oid,
                        order.eid,
                        order.symbol,
                        order.side.value if order.side is not None else None,
                        order.type.value if order.type is not None else None,
                        str(order.amount),
                        order.price or order.average,
                        order.status.value,
                        str(order.fee) if order.fee is not None else None,
                        order.fee_currency,
                        self._encode(order),
                    ),
                )
            await self._db_async.commit()

    async def sync_algo_orders(self, mem_algo_orders: Dict[str, AlgoOrder]) -> None:
        async with self._db_async.cursor() as cursor:
            for oid, algo_order in mem_algo_orders.copy().items():
                await cursor.execute(
                    f"INSERT OR REPLACE INTO {self.table_prefix}_algo_orders "
                    "(timestamp, oid, symbol, data) VALUES (?, ?, ?, ?)",
                    (
                        algo_order.timestamp,
                        oid,
                        algo_order.symbol,
                        self._encode(algo_order),
                    ),
                )
            await self._db_async.commit()

    async def sync_positions(self, mem_positions: Dict[str, Position]) -> None:
        async with self._db_async.cursor() as cursor:
            await cursor.execute(f"SELECT symbol FROM {self.table_prefix}_positions")
            db_positions = {row[0] for row in await cursor.fetchall()}

            positions_to_delete = db_positions - set(mem_positions.keys())
            if positions_to_delete:
                await cursor.executemany(
                    f"DELETE FROM {self.table_prefix}_positions WHERE symbol = ?",
                    [(symbol,) for symbol in positions_to_delete],
                )
                self._log.debug(
                    f"Deleted {len(positions_to_delete)} stale positions from database"
                )

            for symbol, position in mem_positions.copy().items():
                await cursor.execute(
                    f"INSERT OR REPLACE INTO {self.table_prefix}_positions "
                    "(symbol, exchange, side, amount, data) VALUES (?, ?, ?, ?, ?)",
                    (
                        symbol,
                        position.exchange.value,
                        position.side.value if position.side else "FLAT",
                        str(position.amount),
                        self._encode(position),
                    ),
                )
            await self._db_async.commit()

    async def sync_open_orders(
        self,
        mem_open_orders: Dict[ExchangeType, Set[str]],
        mem_orders: Dict[str, Order],
    ) -> None:
        async with self._db_async.cursor() as cursor:
            await cursor.execute(f"DELETE FROM {self.table_prefix}_open_orders")

            for exchange, oids in mem_open_orders.copy().items():
                for oid in oids.copy():
                    order = mem_orders.get(oid)
                    if order:
                        await cursor.execute(
                            f"INSERT INTO {self.table_prefix}_open_orders "
                            "(oid, eid, exchange, symbol) VALUES (?, ?, ?, ?)",
                            (oid, order.eid, exchange.value, order.symbol),
                        )
            await self._db_async.commit()

    async def sync_balances(
        self, mem_account_balance: Dict[AccountType, AccountBalance]
    ) -> None:
        async with self._db_async.cursor() as cursor:
            for account_type, balance in mem_account_balance.copy().items():
                for asset, amount in balance.balances.items():
                    await cursor.execute(
                        f"INSERT OR REPLACE INTO {self.table_prefix}_balances "
                        "(asset, account_type, free, locked) VALUES (?, ?, ?, ?)",
                        (
                            asset,
                            account_type.value,
                            str(amount.free),
                            str(amount.locked),
                        ),
                    )
            await self._db_async.commit()

    def get_order(
        self,
        oid: str,
        mem_orders: Dict[str, Order],
        mem_algo_orders: Dict[str, AlgoOrder],
    ) -> Optional[Order | AlgoOrder]:
        if order := mem_orders.get(oid):
            return order
        if algo_order := mem_algo_orders.get(oid):
            return algo_order

        cursor = self._get_cursor()
        try:
            # Try regular orders first
            cursor.execute(
                f"SELECT data FROM {self.table_prefix}_orders WHERE oid = ?",
                (oid,),
            )
            if row := cursor.fetchone():
                order = self._decode(row[0], Order)
                mem_orders[oid] = order
                return order

            # Try algo orders
            cursor.execute(
                f"SELECT data FROM {self.table_prefix}_algo_orders WHERE oid = ?",
                (oid,),
            )
            if row := cursor.fetchone():
                algo_order = self._decode(row[0], AlgoOrder)
                mem_algo_orders[oid] = algo_order
                return algo_order

            return None

        except sqlite3.Error as e:
            self._log.error(f"Error getting order from SQLite: {e}")
            return None

    def get_symbol_orders(self, symbol: str) -> Set[str]:
        cursor = self._get_cursor()
        cursor.execute(
            f"SELECT oid FROM {self.table_prefix}_orders WHERE symbol = ?",
            (symbol,),
        )
        return {row[0] for row in cursor.fetchall()}

    def get_all_positions(self, exchange_id: ExchangeType) -> Dict[str, Position]:
        positions = {}
        cursor = self._get_cursor()
        cursor.execute(
            f"SELECT symbol, data FROM {self.table_prefix}_positions WHERE exchange = ?",
            (exchange_id.value,),
        )
        for row in cursor.fetchall():
            position = self._decode(row[1], Position)
            if position.side:
                positions[position.symbol] = position
        return positions

    def get_all_balances(self, account_type: AccountType) -> List[Balance]:
        from decimal import Decimal

        balances = []
        cursor = self._get_cursor()
        cursor.execute(
            f"SELECT asset, free, locked FROM {self.table_prefix}_balances WHERE account_type = ?",
            (account_type.value,),
        )
        for row in cursor.fetchall():
            balances.append(
                Balance(asset=row[0], free=Decimal(row[1]), locked=Decimal(row[2]))
            )
        return balances

    async def sync_params(self, mem_params: Dict[str, Any]) -> None:
        import msgspec

        async with self._db_async.cursor() as cursor:
            for key, value in mem_params.copy().items():
                try:
                    serialized_value = self._encode_param(value)
                    await cursor.execute(
                        f"INSERT OR REPLACE INTO {self.table_prefix}_params "
                        "(key, value) VALUES (?, ?)",
                        (key, serialized_value),
                    )
                except msgspec.EncodeError as e:
                    self._log.error(f"Error serializing parameter {key}: {e}")
                except sqlite3.Error as e:
                    self._log.error(f"Error storing parameter {key}: {e}")
            await self._db_async.commit()

    def get_param(self, key: str, default: Any = None) -> Any:
        import msgspec

        try:
            cursor = self._get_cursor()
            cursor.execute(
                f"SELECT value FROM {self.table_prefix}_params WHERE key = ?", (key,)
            )
            if row := cursor.fetchone():
                try:
                    return self._decode_param(row[0])
                except msgspec.DecodeError as e:
                    self._log.error(f"Error deserializing parameter {key}: {e}")
                    return default
            return default
        except sqlite3.Error as e:
            self._log.error(f"Error getting parameter {key}: {e}")
            return default

    def get_all_params(self) -> Dict[str, Any]:
        import msgspec

        params = {}
        try:
            cursor = self._get_cursor()
            cursor.execute(f"SELECT key, value FROM {self.table_prefix}_params")
            for row in cursor.fetchall():
                try:
                    params[row[0]] = self._decode_param(row[1])
                except msgspec.DecodeError as e:
                    self._log.error(f"Error deserializing parameter {row[0]}: {e}")
        except sqlite3.Error as e:
            self._log.error(f"Error getting all parameters: {e}")
        return params



import asyncio
from typing import Dict, List
from decimal import Decimal
from nexustrader.constants import AccountType, SubmitType
from nexustrader.schema import OrderSubmit, InstrumentId
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.bybit import BybitAccountType
from nexustrader.exchange.bybit.schema import BybitMarket
from nexustrader.base import ExecutionManagementSystem


class BybitExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, BybitMarket]

    def __init__(
        self,
        market: Dict[str, BybitMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        registry: OrderRegistry,
        is_mock: bool = False,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            clock=clock,
            task_manager=task_manager,
            registry=registry,
            is_mock=is_mock,
        )
        self._bybit_account_type: BybitAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, BybitAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()

    def _set_account_type(self):
        account_types = self._private_connectors.keys()
        self._bybit_account_type = (
            BybitAccountType.UNIFIED_TESTNET
            if BybitAccountType.UNIFIED_TESTNET in account_types
            else BybitAccountType.UNIFIED
        )

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        if self._is_mock:
            if instrument_id.is_spot:
                return BybitAccountType.SPOT_MOCK
            elif instrument_id.is_linear:
                return BybitAccountType.LINEAR_MOCK
            elif instrument_id.is_inverse:
                return BybitAccountType.INVERSE_MOCK
        else:
            return self._bybit_account_type

    def _submit_order(
        self,
        order: OrderSubmit | List[OrderSubmit],
        submit_type: SubmitType,
        account_type: AccountType | None = None,
    ):
        if isinstance(order, list):
            if not account_type:
                account_type = self._instrument_id_to_account_type(
                    order[0].instrument_id
                )

            # Split batch orders into chunks of 20
            for i in range(0, len(order), 20):
                batch = order[i : i + 20]
                self._order_submit_queues[account_type].put_nowait((batch, submit_type))
        else:
            if not account_type:
                account_type = self._instrument_id_to_account_type(order.instrument_id)
            self._order_submit_queues[account_type].put_nowait((order, submit_type))

    def _get_min_order_amount(
        self, symbol: str, market: BybitMarket, px: float
    ) -> Decimal:
        min_order_qty = float(market.info.lotSizeFilter.minOrderQty)
        min_order_amt = float(
            market.info.lotSizeFilter.minOrderAmt
            or market.info.lotSizeFilter.minNotionalValue
        )
        min_order_amount = max(min_order_amt * 1.02 / px, min_order_qty)
        min_order_amount = self._amount_to_precision(
            symbol, min_order_amount, mode="ceil"
        )
        return min_order_amount



import signal
import asyncio
import uuid
import warnings
from typing import Callable, Coroutine, Any, TypeVar, Union
from typing import Dict, List
from dataclasses import dataclass
from nexustrader.core.nautilius_core import LiveClock, Logger
from nexustrader.schema import (
    Kline,
    BookL1,
    Trade,
    FundingRate,
    BookL2,
    IndexPrice,
    MarkPrice,
)

T = TypeVar("T")
InputDataType = Union[
    Kline, BookL1, Trade, FundingRate, BookL2, IndexPrice, IndexPrice, MarkPrice
]


class OidGen:
    __slots__ = ("_shard", "_last_ms", "_seq", "_clock")

    def __init__(self, clock: LiveClock):
        self._shard = self._generate_shard()
        self._last_ms = 0
        self._seq = 0
        self._clock = clock

    def _generate_shard(self) -> int:
        uuid_int = uuid.uuid4().int
        shard = (uuid_int % 9999) + 1
        return shard

    @property
    def oid(self) -> str:
        ms = self._clock.timestamp_ms()
        if ms == self._last_ms:
            self._seq = (self._seq + 1) % 1000
            if self._seq == 0:
                while True:
                    ms2 = self._clock.timestamp_ms()
                    if ms2 > ms:
                        ms = ms2
                        break
        else:
            self._seq = 0
        self._last_ms = ms
        return f"{ms:013d}{self._seq:03d}{self._shard:04d}"

    def get_shard(self) -> int:
        return self._shard


def is_redis_available() -> bool:
    """Check if Redis dependencies and server are available"""
    try:
        # Check if Redis Python packages are installed
        import redis
        import socket
        from nexustrader.constants import get_redis_config

        # Check if Redis server is accessible
        try:
            # Get Redis config and test connection
            redis_config = get_redis_config()
            client = redis.Redis(**redis_config)
            client.ping()  # This will raise an exception if Redis is not accessible
            client.close()
            return True
        except (redis.ConnectionError, redis.TimeoutError, socket.error, Exception):
            return False

    except ImportError:
        return False


def get_redis_client_if_available():
    import redis

    """Get Redis client if available, otherwise return None"""
    if not is_redis_available():
        return None

    try:
        from nexustrader.constants import get_redis_config

        redis_config = get_redis_config()
        return redis.Redis(**redis_config)
    except Exception:
        return None


@dataclass
class RateLimit:
    """
    max_rate: Allow up to max_rate / time_period acquisitions before blocking.
    time_period: Time period in seconds.
    """

    max_rate: float
    time_period: float = 60


class TaskManager:
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        enable_signal_handlers: bool = True,
        cancel_timeout: float | None = 5.0,
    ):
        self._log = Logger(name=type(self).__name__)
        self._tasks: Dict[str, asyncio.Task] = {}
        self._shutdown_event = asyncio.Event()
        self._loop = loop
        self._cancel_timeout = cancel_timeout
        if enable_signal_handlers:
            self._setup_signal_handlers()

    def _setup_signal_handlers(self):
        try:
            for sig in (signal.SIGINT, signal.SIGTERM):
                self._loop.add_signal_handler(
                    sig, lambda: self.create_task(self._shutdown())
                )
        except NotImplementedError:
            warnings.warn("Signal handlers not supported on this platform")

    async def _shutdown(self):
        self._shutdown_event.set()
        self._log.debug("Shutdown signal received, cleaning up...")

    def create_task(self, coro: asyncio.coroutines, name: str = None) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        self._tasks[task.get_name()] = task
        task.add_done_callback(self._handle_task_done)
        return task

    def run_sync(self, coro: Coroutine[Any, Any, T]) -> T:
        """
        Run an async coroutine in a synchronous context.

        Args:
            coro: The coroutine to run

        Returns:
            The result of the coroutine

        Raises:
            RuntimeError: If the event loop is not running and cannot be started
            Exception: Any exception raised by the coroutine
        """
        try:
            if self._loop.is_running():
                future = asyncio.run_coroutine_threadsafe(coro, self._loop)
                return future.result()
            else:
                if self._loop.is_closed():
                    raise RuntimeError("Event loop is closed")
                return self._loop.run_until_complete(coro)
        except asyncio.CancelledError:
            raise RuntimeError("Coroutine was cancelled")
        except Exception as e:
            self._log.error(f"Error running coroutine: {e}")
            raise

    def cancel_task(self, name: str) -> bool:
        if name in self._tasks:
            self._tasks[name].cancel()
            return True
        return False

    def _handle_task_done(self, task: asyncio.Task):
        try:
            name = task.get_name()
            self._tasks.pop(name, None)
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self._log.error(f"Error during task done: {e}")
            raise

    async def wait(self):
        try:
            if self._tasks:
                await self._shutdown_event.wait()
        except Exception as e:
            self._log.error(f"Error during wait: {e}")
            raise

    async def cancel(self, timeout: float | None = None):
        try:
            # Snapshot tasks to avoid dict mutation during iteration
            tasks = [t for t in self._tasks.values() if not t.done()]

            for task in tasks:
                task.cancel()

            if tasks:
                gather_coro = asyncio.gather(*tasks, return_exceptions=True)
                try:
                    effective_timeout = (
                        self._cancel_timeout if timeout is None else timeout
                    )
                    if effective_timeout is not None:
                        results = await asyncio.wait_for(
                            gather_coro, timeout=effective_timeout
                        )
                    else:
                        results = await gather_coro
                except asyncio.TimeoutError:
                    self._log.warning(
                        "Cancellation timed out; some tasks may still be running"
                    )
                else:
                    for result in results:
                        if isinstance(result, Exception) and not isinstance(
                            result, asyncio.CancelledError
                        ):
                            self._log.error(
                                f"Task failed during cancellation: {result}"
                            )

        except Exception as e:
            self._log.error(f"Error during cancellation: {e}")
            raise
        finally:
            self._tasks.clear()


# class Clock:
#     def __init__(self, tick_size: float = 1.0):
#         """
#         :param tick_size_s: Time interval of each tick in seconds (supports sub-second precision).
#         """
#         self._tick_size = tick_size  # Tick size in seconds
#         self._current_tick = (time.time() // self._tick_size) * self._tick_size
#         self._clock = LiveClock()
#         self._tick_callbacks: List[Callable[[float], None]] = []
#         self._started = False

#     @property
#     def tick_size(self) -> float:
#         return self._tick_size

#     @property
#     def current_timestamp(self) -> float:
#         return self._clock.timestamp()

#     def add_tick_callback(self, callback: Callable[[float], None]):
#         """
#         Register a callback to be called on each tick.
#         :param callback: Function to be called with current_tick as argument.
#         """
#         self._tick_callbacks.append(callback)

#     async def run(self):
#         if self._started:
#             raise RuntimeError("Clock is already running.")
#         self._started = True
#         while True:
#             now = time.time()
#             next_tick_time = self._current_tick + self._tick_size
#             sleep_duration = next_tick_time - now
#             if sleep_duration > 0:
#                 await asyncio.sleep(sleep_duration)
#             else:
#                 # If we're behind schedule, skip to the next tick to prevent drift
#                 next_tick_time = now
#             self._current_tick = next_tick_time
#             for callback in self._tick_callbacks:
#                 if asyncio.iscoroutinefunction(callback):
#                     await callback(self.current_timestamp)
#                 else:
#                     callback(self.current_timestamp)


class ZeroMQSignalRecv:
    def __init__(self, config, callback: Callable, task_manager: TaskManager):
        self._socket = config.socket
        self._callback = callback
        self._task_manager = task_manager

    async def _recv(self):
        while True:
            date = await self._socket.recv()
            if asyncio.iscoroutinefunction(self._callback):
                await self._callback(date)
            else:
                self._callback(date)

    async def start(self):
        self._task_manager.create_task(self._recv())


# class MovingAverage:
#     """
#     Calculate moving median or mean using a sliding window.

#     Args:
#         length: Length of the sliding window
#         method: 'median' or 'mean' calculation method
#     """

#     def __init__(self, length: int, method: Literal["median", "mean"] = "mean"):
#         if method not in ["median", "mean"]:
#             raise ValueError("method must be either 'median' or 'mean'")

#         self._length = length
#         self._method = method
#         self._window = deque(maxlen=length)
#         self._calc_func = median if method == "median" else mean

#     def input(self, value: float) -> float | None:
#         """
#         Input a new value and return the current median/mean.

#         Args:
#             value: New value to add to sliding window

#         Returns:
#             Current median/mean value, or None if window not filled
#         """
#         self._window.append(value)

#         if len(self._window) < self._length:
#             return None

#         return self._calc_func(self._window)


class DataReady:
    def __init__(
        self,
        symbols: List[str],
        name: str,
        clock: LiveClock,
        timeout: int = 60,
        permanently_ready: bool = False,
    ):
        """
        Initialize DataReady class

        Args:
            symbols: symbols list need to monitor
            timeout: timeout in seconds
        """
        self._log = Logger(name=type(self).__name__)

        # Optimization: Store symbols as a set for faster lookups if needed,
        # but dict is fine for tracking status.
        self._symbols_status = {symbol: False for symbol in symbols}
        self._total_symbols = len(symbols)
        self._ready_symbols_count = 0

        self._timeout_ms = timeout * 1000  # Store timeout in ms
        self._clock = clock
        self._first_data_time: int | None = None
        self._name = name
        # Optimization: A flag to indicate that the "ready" state is final
        # (either all symbols received or timed out).
        self._is_permanently_ready: bool = permanently_ready

        if not symbols:  # If no symbols to monitor, it's ready immediately (or upon first data for timestamp)
            # We'll consider it ready once _first_data_time is set, which requires at least one input call
            # Or, we can decide it's ready right away if that's the desired behavior.
            # For now, let's assume if symbols is empty, it becomes ready when `ready` is first checked
            # after _first_data_time is set, or on timeout.
            # Alternatively, one could set self._is_permanently_ready = True here.
            # The current logic for _ready_symbols_count == _total_symbols (0 == 0) will handle this.
            self._is_permanently_ready = True

    def input(self, data: InputDataType) -> None:
        """
        Input data, update the status of the corresponding symbol

        Args:
            data: data object with symbol attribute
        """
        # Optimization: If already permanently ready, do nothing.
        if self._is_permanently_ready:
            return

        symbol = data.symbol

        if self._first_data_time is None:
            self._first_data_time = self._clock.timestamp_ms()

        # Process only if the symbol is one we are monitoring and it's not yet marked ready.
        if symbol in self._symbols_status and not self._symbols_status[symbol]:
            self._symbols_status[symbol] = True
            self._ready_symbols_count += 1

            # Optimization: Check if all symbols are now ready.
            if self._ready_symbols_count == self._total_symbols:
                self._log.debug(
                    f"All {self._total_symbols} symbols received. {self._name} is ready."
                )
                self._is_permanently_ready = True
                # No need to call self.ready here, the flag is enough.

    def add_symbols(self, symbols: str | list[str]) -> None:
        """
        Add a new symbol to monitor

        Args:
            symbol: symbol to add
        """
        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            if symbol not in self._symbols_status:
                self._symbols_status[symbol] = False
                self._total_symbols += 1
                # If we were already permanently ready, adding a new symbol means we are no longer ready.
                if self._is_permanently_ready:
                    self._is_permanently_ready = False

    @property
    def ready(self) -> bool:
        """
        Check if all data is ready or if it has timed out

        Returns:
            bool: if all data is ready or timed out, return True
        """
        # Optimization: If already determined to be permanently ready, return True.
        if self._is_permanently_ready:
            return True

        if self._first_data_time is None:
            # No data received yet, so not ready (unless symbols list was empty and we decided that's ready from init)
            return False  # Or handle empty symbols list case from __init__

        # Check if all symbols have been received (could have been set by input)
        if self._ready_symbols_count == self._total_symbols:
            # This ensures that even if input didn't set _is_permanently_ready
            # (e.g. if ready is called before next input after last symbol arrived),
            # it's correctly identified.
            if (
                not self._is_permanently_ready
            ):  # To avoid repeated logging if already logged by input
                self._log.debug(
                    f"All {self._total_symbols} symbols confirmed ready by property access."
                )
            self._is_permanently_ready = True
            return True

        # Check for timeout
        if self._clock.timestamp_ms() - self._first_data_time > self._timeout_ms:
            # Only issue warning if not all symbols were ready before timeout
            if self._ready_symbols_count < self._total_symbols:
                not_ready_symbols = [
                    symbol
                    for symbol, status in self._symbols_status.items()
                    if not status
                ]
                if not_ready_symbols:  # Should always be true if count < total
                    warnings.warn(
                        f"Data receiving timed out. Timeout: {self._timeout_ms / 1000}s. "
                        f"Received {self._ready_symbols_count}/{self._total_symbols} symbols. "
                        f"The following symbols are not ready: {', '.join(not_ready_symbols)}"
                    )
            else:  # All symbols were received, but timeout check ran.
                self._log.debug(
                    f"Timeout checked, but all symbols were already received. {self._name} is ready."
                )

            self._is_permanently_ready = True  # Timed out, so state is now final.
            return True

        return False  # Not all symbols ready and not timed out yet.



import msgspec
import asyncio
import threading
import re
import redis
from typing import Dict, Set, Type, List, Optional, Any
from collections import defaultdict
from returns.maybe import maybe
from pathlib import Path

from nexustrader.schema import (
    Order,
    Position,
    ExchangeType,
    Kline,
    BookL1,
    Trade,
    AlgoOrder,
    AccountBalance,
    Balance,
    FundingRate,
    IndexPrice,
    MarkPrice,
    BookL2,
)
from nexustrader.constants import STATUS_TRANSITIONS, AccountType, KlineInterval
from nexustrader.core.entity import TaskManager, get_redis_client_if_available
from nexustrader.core.nautilius_core import LiveClock, MessageBus, Logger
from nexustrader.constants import StorageType, ParamBackend
from nexustrader.backends import SQLiteBackend, PostgreSQLBackend


class AsyncCache:
    _backend: SQLiteBackend | PostgreSQLBackend

    def __init__(
        self,
        strategy_id: str,
        user_id: str,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        storage_backend: StorageType = StorageType.SQLITE,
        db_path: str = ".keys/cache.db",
        sync_interval: int = 60,  # seconds
        expired_time: int = 3600,  # seconds
    ):
        parent_dir = Path(db_path).parent
        if not parent_dir.exists():
            parent_dir.mkdir(parents=True, exist_ok=True)

        self.strategy_id = strategy_id
        self.user_id = user_id
        self._storage_backend = storage_backend
        self._db_path = db_path

        self._log = Logger(name=type(self).__name__)
        self._clock = clock

        # in-memory save
        self._mem_orders: Dict[str, Order] = {}  # oid -> Order
        self._mem_algo_orders: Dict[str, AlgoOrder] = {}  # oid -> AlgoOrder
        self._mem_open_orders: Dict[ExchangeType, Set[str]] = defaultdict(
            set
        )  # exchange_id -> set(oid)
        self._mem_symbol_open_orders: Dict[str, Set[str]] = defaultdict(
            set
        )  # symbol -> set(oid)
        self._mem_symbol_orders: Dict[str, Set[str]] = defaultdict(
            set
        )  # symbol -> set(oid)
        self._mem_positions: Dict[str, Position] = {}  # symbol -> Position
        self._mem_account_balance: Dict[AccountType, AccountBalance] = defaultdict(
            AccountBalance
        )
        self._mem_params: Dict[str, Any] = {}  # params cache
        self._cancel_intent_oids: Set[str] = (
            set()
        )  # oids currently pending cancel intent

        # set params
        self._sync_interval = sync_interval  # sync interval
        self._expired_time = expired_time  # expire time
        self._task_manager = task_manager

        self._kline_cache: Dict[str, Kline] = {}
        self._bookl1_cache: Dict[str, BookL1] = {}
        self._trade_cache: Dict[str, Trade] = {}
        self._bookl2_cache: Dict[str, BookL2] = {}
        self._funding_rate_cache: Dict[str, FundingRate] = {}
        self._index_price_cache: Dict[str, IndexPrice] = {}
        self._mark_price_cache: Dict[str, MarkPrice] = {}

        self._msgbus = msgbus
        self._msgbus.subscribe(topic="kline", handler=self._update_kline_cache)
        self._msgbus.subscribe(topic="bookl1", handler=self._update_bookl1_cache)
        self._msgbus.subscribe(topic="trade", handler=self._update_trade_cache)
        self._msgbus.subscribe(topic="bookl2", handler=self._update_bookl2_cache)
        self._msgbus.subscribe(
            topic="funding_rate", handler=self._update_funding_rate_cache
        )
        self._msgbus.subscribe(
            topic="index_price", handler=self._update_index_price_cache
        )
        self._msgbus.subscribe(
            topic="mark_price", handler=self._update_mark_price_cache
        )

        self._storage_initialized = False
        self._table_prefix = self.safe_table_name(f"{self.strategy_id}_{self.user_id}")
        self._backend = None

        self._position_lock = threading.RLock()  # Lock for position updates
        self._order_lock = threading.RLock()  # Lock for order updates
        self._balance_lock = threading.RLock()  # Lock for balance updates

        # Redis client for parameter storage (lazy initialization)
        self._redis_client = None
        self._redis_client_initialized = False

    ################# # base functions ####################

    @staticmethod
    def safe_table_name(name: str) -> str:
        name = re.sub(r"[^a-zA-Z0-9_]", "_", name)
        return name.lower()

    def _encode(self, obj: Order | Position | AlgoOrder) -> bytes:
        return msgspec.json.encode(obj)

    def _decode(
        self, data: bytes, obj_type: Type[Order | Position | AlgoOrder]
    ) -> Order | Position | AlgoOrder:
        return msgspec.json.decode(data, type=obj_type)

    async def _init_storage(self):
        """Initialize the storage backend"""
        if self._storage_backend == StorageType.SQLITE:
            self._backend = SQLiteBackend(
                strategy_id=self.strategy_id,
                user_id=self.user_id,
                table_prefix=self._table_prefix,
                log=self._log,
                db_path=self._db_path,
            )
        elif self._storage_backend == StorageType.POSTGRESQL:
            self._backend = PostgreSQLBackend(
                strategy_id=self.strategy_id,
                user_id=self.user_id,
                table_prefix=self._table_prefix,
                log=self._log,
            )

        assert self._backend is not None

        await self._backend.start()
        self._storage_initialized = True

    async def _load_params_from_db(self):
        """Load existing parameters from database"""
        try:
            existing_params = self._backend.get_all_params()
            self._mem_params.update(existing_params)
            if existing_params:
                self._log.debug(
                    f"Loaded {len(existing_params)} parameters from database"
                )
        except Exception as e:
            self._log.error(f"Error loading parameters from database: {e}")

    async def start(self):
        """Start the cache"""
        await self._init_storage()
        # Load existing parameters from database
        await self._load_params_from_db()
        self._task_manager.create_task(self._periodic_sync())

    async def _periodic_sync(self):
        """Periodically sync the cache"""
        while True:
            await self._backend.sync_orders(self._mem_orders)
            await self._backend.sync_algo_orders(self._mem_algo_orders)
            await self._backend.sync_positions(self._mem_positions)
            await self._backend.sync_open_orders(
                self._mem_open_orders, self._mem_orders
            )
            await self._backend.sync_balances(self._mem_account_balance)
            await self._backend.sync_params(self._mem_params)
            self._cleanup_expired_data()
            await asyncio.sleep(self._sync_interval)

    async def sync_orders(self):
        with self._order_lock:
            await self._backend.sync_orders(self._mem_orders)

    async def sync_algo_orders(self):
        with self._order_lock:
            await self._backend.sync_algo_orders(self._mem_algo_orders)

    async def sync_positions(self):
        with self._position_lock:
            await self._backend.sync_positions(self._mem_positions)

    async def sync_open_orders(self):
        with self._order_lock:
            await self._backend.sync_open_orders(
                self._mem_open_orders, self._mem_orders
            )

    async def sync_balances(self):
        with self._balance_lock:
            await self._backend.sync_balances(self._mem_account_balance)

    async def sync_params(self):
        await self._backend.sync_params(self._mem_params)

    def _cleanup_expired_data(self):
        """Cleanup expired data"""
        current_time = self._clock.timestamp_ms()
        expire_before = current_time - self._expired_time * 1000

        with self._order_lock:
            expired_orders = []
            for oid, order in self._mem_orders.copy().items():
                if order.timestamp < expire_before:
                    expired_orders.append(oid)

                    if not order.is_closed:
                        self._log.warning(f"order {oid} is not closed, but expired")

            for oid in expired_orders:
                del self._mem_orders[oid]
                self._log.debug(f"removing order {oid} from memory")
                for symbol, order_set in self._mem_symbol_orders.copy().items():
                    self._log.debug(f"removing order {oid} from symbol {symbol}")
                    order_set.discard(oid)

            expired_algo_orders = [
                oid
                for oid, algo_order in self._mem_algo_orders.copy().items()
                if algo_order.timestamp < expire_before
            ]
            for oid in expired_algo_orders:
                del self._mem_algo_orders[oid]
                self._log.debug(f"removing algo order {oid} from memory")

    async def close(self):
        """关闭缓存"""
        if self._storage_initialized and self._backend:
            try:
                await self._backend.sync_orders(self._mem_orders)
                await self._backend.sync_algo_orders(self._mem_algo_orders)
                await self._backend.sync_positions(self._mem_positions)
                await self._backend.sync_open_orders(
                    self._mem_open_orders, self._mem_orders
                )
                await self._backend.sync_balances(self._mem_account_balance)
                await self._backend.sync_params(self._mem_params)
            except Exception as e:
                # Never let cache close crash shutdown; log and continue
                self._log.error(f"Error closing cache (sync phase): {e}")
            finally:
                try:
                    await self._backend.close()
                except Exception as e:
                    self._log.error(f"Error closing storage backend: {e}")

    ################ # cache public data  ###################

    def _update_kline_cache(self, kline: Kline):
        key = f"{kline.symbol}-{kline.interval.value}"
        self._kline_cache[key] = kline

    def _update_bookl1_cache(self, bookl1: BookL1):
        self._bookl1_cache[bookl1.symbol] = bookl1

    def _update_trade_cache(self, trade: Trade):
        self._trade_cache[trade.symbol] = trade

    def _update_bookl2_cache(self, bookl2: BookL2):
        self._bookl2_cache[bookl2.symbol] = bookl2

    def _update_funding_rate_cache(self, funding_rate: FundingRate):
        self._funding_rate_cache[funding_rate.symbol] = funding_rate

    def _update_index_price_cache(self, index_price: IndexPrice):
        self._index_price_cache[index_price.symbol] = index_price

    def _update_mark_price_cache(self, mark_price: MarkPrice):
        self._mark_price_cache[mark_price.symbol] = mark_price

    def kline(self, symbol: str, interval: KlineInterval) -> Optional[Kline]:
        """
        Retrieve a Kline object from the cache by symbol.

        :param symbol: The symbol of the Kline to retrieve.
        :return: The Kline object if found, otherwise None.
        """
        key = f"{symbol}-{interval.value}"
        return self._kline_cache.get(key, None)

    def bookl1(self, symbol: str) -> Optional[BookL1]:
        """
        Retrieve a BookL1 object from the cache by symbol.

        :param symbol: The symbol of the BookL1 to retrieve.
        :return: The BookL1 object if found, otherwise None.
        """
        return self._bookl1_cache.get(symbol, None)

    def bookl2(self, symbol: str) -> Optional[BookL2]:
        """
        Retrieve a BookL2 object from the cache by symbol.
        """
        return self._bookl2_cache.get(symbol, None)

    def trade(self, symbol: str) -> Optional[Trade]:
        """
        Retrieve a Trade object from the cache by symbol.

        :param symbol: The symbol of the Trade to retrieve.
        :return: The Trade object if found, otherwise None.
        """
        return self._trade_cache.get(symbol, None)

    def funding_rate(self, symbol: str) -> Optional[FundingRate]:
        """
        Retrieve a FundingRate object from the cache by symbol.
        """
        return self._funding_rate_cache.get(symbol, None)

    def index_price(self, symbol: str) -> Optional[IndexPrice]:
        """
        Retrieve an IndexPrice object from the cache by symbol.
        """
        return self._index_price_cache.get(symbol, None)

    def mark_price(self, symbol: str) -> Optional[MarkPrice]:
        """
        Retrieve a MarkPrice object from the cache by symbol.
        """
        return self._mark_price_cache.get(symbol, None)

    ################ # cache private data  ###################

    def _check_status_transition(self, order: Order):
        previous_order = self._mem_orders.get(order.oid)
        if not previous_order:
            return True

        if order.status not in STATUS_TRANSITIONS[previous_order.status]:
            self._log.warning(
                f"Order id: {order.oid} Invalid status transition: {previous_order.status} -> {order.status}"
            )
            return False

        return True

    def _apply_position(self, position: Position):
        with self._position_lock:
            if position.is_closed:
                self._mem_positions.pop(position.symbol, None)
            else:
                self._mem_positions[position.symbol] = position

    def _apply_balance(self, account_type: AccountType, balances: List[Balance]):
        with self._balance_lock:
            self._mem_account_balance[account_type]._apply(balances)

    def get_balance(self, account_type: AccountType) -> AccountBalance:
        with self._balance_lock:
            return self._mem_account_balance[account_type]

    @maybe
    def get_position(self, symbol: str) -> Optional[Position]:
        with self._position_lock:
            if position := self._mem_positions.get(symbol, None):
                return position

    def get_all_positions(
        self, exchange: Optional[ExchangeType] = None
    ) -> Dict[str, Position]:
        with self._position_lock:
            positions = {
                symbol: position
                for symbol, position in self._mem_positions.copy().items()
                if (
                    (exchange is None or position.exchange == exchange)
                    and position.is_opened
                )
            }
            return positions

    def _order_status_update(self, order: Order | AlgoOrder) -> bool:
        with self._order_lock:
            if isinstance(order, AlgoOrder):
                self._mem_algo_orders[order.oid] = order
            else:
                if not self._check_status_transition(order):
                    return False
                self._mem_orders[order.oid] = order

                # Ensure order is tracked in all sets if it's open (handles WebSocket arriving before REST API)
                if not order.is_closed:
                    self._mem_open_orders[order.exchange].add(order.oid)
                    self._mem_symbol_orders[order.symbol].add(order.oid)
                    self._mem_symbol_open_orders[order.symbol].add(order.oid)
                else:
                    self._mem_open_orders[order.exchange].discard(order.oid)
                    self._mem_symbol_open_orders[order.symbol].discard(order.oid)
                    self._cancel_intent_oids.discard(order.oid)
                return True

    def mark_all_cancel_intent(self, symbol: str) -> None:
        with self._order_lock:
            oids = self._mem_symbol_open_orders.get(symbol, set())
            self._cancel_intent_oids.update(oids)

    def mark_cancel_intent(self, oid: str) -> None:
        with self._order_lock:
            self._cancel_intent_oids.add(oid)

    # NOTE: this function is not for user to call, it is for internal use
    def _get_all_balances_from_db(self, account_type: AccountType) -> List[Balance]:
        with self._balance_lock:
            return self._backend.get_all_balances(account_type)

    # NOTE: this function is not for user to call, it is for internal use
    def _get_all_positions_from_db(
        self, exchange_id: ExchangeType
    ) -> Dict[str, Position]:
        return self._backend.get_all_positions(exchange_id)

    @maybe
    def get_order(self, oid: str) -> Optional[Order | AlgoOrder]:
        with self._order_lock:
            return self._backend.get_order(oid, self._mem_orders, self._mem_algo_orders)

    def get_symbol_orders(self, symbol: str, in_mem: bool = True) -> Set[str]:
        """Get all orders for a symbol from memory and storage"""
        with self._order_lock:
            memory_orders = self._mem_symbol_orders.get(symbol, set())
            if not in_mem:
                storage_orders = self._backend.get_symbol_orders(symbol)
                return memory_orders.union(storage_orders)
            return memory_orders

    def get_open_orders(
        self,
        symbol: str | None = None,
        exchange: ExchangeType | None = None,
        *,
        include_canceling: bool = False,
    ) -> Set[str]:
        with self._order_lock:
            if symbol is not None:
                orders = self._mem_symbol_open_orders[symbol].copy()
            elif exchange is not None:
                orders = self._mem_open_orders[exchange].copy()
            else:
                raise ValueError("Either `symbol` or `exchange` must be specified")

            if include_canceling:
                return orders

            return orders.difference(self._cancel_intent_oids)

    ################ # parameter cache  ###################

    def _ensure_redis_client(self) -> None:
        """
        Lazy initialization of Redis client.
        Raises RuntimeError if Redis is not available.
        """
        if not self._redis_client_initialized:
            self._redis_client = get_redis_client_if_available()
            self._redis_client_initialized = True
            if self._redis_client is None:
                raise RuntimeError(
                    "Redis client is not available. Please ensure Redis is running and configured properly."
                )

    def _get_redis_param_key(self, key: str) -> str:
        """Generate Redis key for parameter storage"""
        return f"nexus:param:{self.strategy_id}:{self.user_id}:{key}"

    def get_param(
        self,
        key: str,
        default: Any = None,
        param_backend: ParamBackend = ParamBackend.MEMORY,
    ) -> Any:
        """
        Get a parameter from the cache

        Args:
            key: Parameter key
            default: Default value if key not found
            param_backend: Storage backend (MEMORY or REDIS)

        Returns:
            Parameter value or default

        Raises:
            RuntimeError: If param_backend is REDIS but Redis is not available
        """
        if param_backend == ParamBackend.REDIS:
            self._ensure_redis_client()
            redis_key = self._get_redis_param_key(key)
            value = self._redis_client.get(redis_key)
            if value is None:
                return default
            # Deserialize the value
            return msgspec.json.decode(value)
        else:
            return self._mem_params.get(key, default)

    def set_param(
        self, key: str, value: Any, param_backend: ParamBackend = ParamBackend.MEMORY
    ) -> None:
        """
        Set a parameter in the cache

        Args:
            key: Parameter key
            value: Parameter value
            param_backend: Storage backend (MEMORY or REDIS)

        Raises:
            RuntimeError: If param_backend is REDIS but Redis is not available
        """
        if param_backend == ParamBackend.REDIS:
            self._ensure_redis_client()
            redis_key = self._get_redis_param_key(key)
            # Serialize the value
            serialized_value = msgspec.json.encode(value)
            self._redis_client.set(redis_key, serialized_value)
        else:
            self._mem_params[key] = value

    def get_all_params(
        self, param_backend: ParamBackend = ParamBackend.MEMORY
    ) -> Dict[str, Any]:
        """
        Get all parameters from the cache

        Args:
            param_backend: Storage backend (MEMORY or REDIS)

        Returns:
            Dictionary of all parameters

        Raises:
            RuntimeError: If param_backend is REDIS but Redis is not available
        """
        if param_backend == ParamBackend.REDIS:
            self._ensure_redis_client()
            pattern = f"nexus:param:{self.strategy_id}:{self.user_id}:*"
            params = {}
            # Use SCAN to iterate over keys matching the pattern
            cursor = 0
            while True:
                cursor, keys = self._redis_client.scan(cursor, match=pattern, count=100)
                for redis_key in keys:
                    # Extract the parameter key from the Redis key
                    key = (
                        redis_key.decode()
                        if isinstance(redis_key, bytes)
                        else redis_key
                    )
                    param_key = key.split(":", 4)[-1]  # Get the part after the last ':'
                    value = self._redis_client.get(redis_key)
                    if value is not None:
                        params[param_key] = msgspec.json.decode(value)
                if cursor == 0:
                    break
            return params
        else:
            return self._mem_params.copy()

    def clear_param(
        self,
        key: Optional[str] = None,
        param_backend: ParamBackend = ParamBackend.MEMORY,
    ) -> None:
        """
        Clear parameter(s) from the cache

        Args:
            key: Parameter key to clear (None to clear all)
            param_backend: Storage backend (MEMORY or REDIS)

        Raises:
            RuntimeError: If param_backend is REDIS but Redis is not available
        """
        if param_backend == ParamBackend.REDIS:
            self._ensure_redis_client()
            if key is None:
                # Clear all parameters matching the pattern
                pattern = f"nexus:param:{self.strategy_id}:{self.user_id}:*"
                cursor = 0
                while True:
                    cursor, keys = self._redis_client.scan(
                        cursor, match=pattern, count=100
                    )
                    if keys:
                        self._redis_client.delete(*keys)
                    if cursor == 0:
                        break
            else:
                # Clear specific parameter
                redis_key = self._get_redis_param_key(key)
                self._redis_client.delete(redis_key)
        else:
            if key is None:
                # Clear all parameters
                self._mem_params.clear()
            else:
                # Clear specific parameter
                self._mem_params.pop(key, None)

    ################ # Redis direct access  ###################

    @property
    def redis(self) -> Optional[redis.Redis]:
        """
        Direct access to Redis client for calling any Redis methods.

        Returns:
            Redis client instance

        Raises:
            RuntimeError: If Redis is not available

        Example:
            # Set a key
            cache.redis.set("mykey", "myvalue")

            # Get a key
            value = cache.redis.get("mykey")

            # Use hash operations
            cache.redis.hset("myhash", "field1", "value1")
            cache.redis.hget("myhash", "field1")

            # Use list operations
            cache.redis.lpush("mylist", "item1")
            cache.redis.rpop("mylist")

            # Use set operations
            cache.redis.sadd("myset", "member1")
            cache.redis.smembers("myset")

            # Use sorted set operations
            cache.redis.zadd("myzset", {"member1": 1.0})
            cache.redis.zrange("myzset", 0, -1)

            # Pipeline operations
            pipe = cache.redis.pipeline()
            pipe.set("key1", "value1")
            pipe.set("key2", "value2")
            pipe.execute()

            # All standard Redis commands are available
        """
        self._ensure_redis_client()
        return self._redis_client



from nautilus_trader.common.component import MessageBus
from nautilus_trader.common.component import LiveClock, TimeEvent
from nautilus_trader.model.identifiers import TraderId
from nautilus_trader.core.uuid import UUID4

from nautilus_trader.core import nautilus_pyo3  # noqa
from nautilus_trader.core.nautilus_pyo3 import HttpClient  # noqa
from nautilus_trader.core.nautilus_pyo3 import HttpMethod  # noqa
from nautilus_trader.core.nautilus_pyo3 import HttpResponse  # noqa

# from nautilus_trader.core.nautilus_pyo3 import MessageBus  # noqa
from nautilus_trader.core.nautilus_pyo3 import WebSocketClient  # noqa
from nautilus_trader.core.nautilus_pyo3 import WebSocketClientError  # noqa
from nautilus_trader.core.nautilus_pyo3 import WebSocketConfig  # noqa
from nautilus_trader.core.nautilus_pyo3 import (
    hmac_signature,  # noqa
    rsa_signature,  # noqa
    ed25519_signature,  # noqa
)
from nautilus_trader.common.component import Logger, set_logging_pyo3  # noqa


def setup_nautilus_core(
    trader_id: str,
    level_stdout: str,
    level_file: str | None = None,
    component_levels: dict[str, str] | None = None,
    directory: str | None = None,
    file_name: str | None = None,
    file_format: str | None = None,
    file_rotate: tuple[int, int] | None = None,
    is_colored: bool | None = None,
    is_bypassed: bool | None = None,
    print_config: bool | None = None,
    log_components_only: bool | None = None,
):
    """
    Setup logging for the application.
    """
    clock = LiveClock()
    msgbus = MessageBus(
        trader_id=TraderId(trader_id),
        clock=clock,
    )
    set_logging_pyo3(True)

    instance_id = nautilus_pyo3.UUID4().value
    log_guard = nautilus_pyo3.init_logging(
        trader_id=nautilus_pyo3.TraderId(trader_id),
        instance_id=nautilus_pyo3.UUID4.from_str(instance_id),
        level_stdout=nautilus_pyo3.LogLevel(level_stdout),
        level_file=nautilus_pyo3.LogLevel(level_file) if level_file else None,
        directory=directory,
        file_name=file_name,
        file_format=file_format,
        is_colored=is_colored,
        print_config=print_config,
        component_levels=component_levels,
        file_rotate=file_rotate,
        is_bypassed=is_bypassed,
        log_components_only=log_components_only,
    )

    return log_guard, msgbus, clock


def usage():
    import time

    print(UUID4().value)
    print(UUID4().value)
    print(UUID4().value)

    uuid_to_order_id = {}

    uuid = UUID4()

    order_id = "123456"

    uuid_to_order_id[uuid] = order_id

    print(uuid_to_order_id)

    clock = LiveClock()
    print(clock.timestamp())
    print(type(clock.timestamp_ms()))

    print(clock.utc_now().isoformat(timespec="milliseconds").replace("+00:00", "Z"))

    def handler1(msg):
        print(f"[{clock.timestamp_ns()}] Received message: {msg} - handler1")

    def handler2(msg):
        print(f"[{clock.timestamp_ns()}] Received message: {msg} - handler2")

    def handler3(msg):
        print(f"[{clock.timestamp_ns()}] Received message: {msg} - handler3")

    log_guard, msgbus, clock = setup_nautilus_core(
        trader_id="TESTER-001",
        level_stdout="DEBUG",
        component_levels={
            "logger1": "DEBUG",
            "logger2": "INFO",
        },
        log_components_only=True,
    )

    log1 = Logger("logger1")
    log2 = Logger("logger2")
    log1.debug("This is a debug msg")
    log1.info("This is a info msg")
    log2.debug("This is a debug msg")
    log2.info("This is a info msg")

    # msgbus.subscribe(topic="order", handler=handler1)
    # msgbus.subscribe(topic="order", handler=handler2)
    # msgbus.subscribe(topic="order", handler=handler3)

    # try:
    #     while True:
    #         msgbus.publish(topic="order", msg="hello")
    #         time.sleep(1)
    # except KeyboardInterrupt:
    #     print("Exiting...")

    # print("done")
    # from datetime import timedelta, datetime, timezone

    # count = 0
    # name = "TEST_TIMER 111"

    # def count_handler(event: TimeEvent):
    #     nonlocal count
    #     count += 1

    #     print(
    #         f"[{clock.utc_now()}] {event.ts_event} {event.ts_init} {clock.timestamp_ns() - clock.next_time_ns(name)} {event.ts_event - clock.next_time_ns(name)}"
    #     )

    # # clock.register_default_handler(count_handler)

    # interval = timedelta(milliseconds=1000)
    # start_time = (datetime.now(tz=timezone.utc) + timedelta(seconds=1)).replace(
    #     microsecond=0
    # )
    # clock.set_timer(
    #     name=name,
    #     interval=interval,
    #     start_time=start_time,
    #     stop_time=None,
    #     callback=count_handler,
    # )

    # time.sleep(10000)


if __name__ == "__main__":
    usage()



from nexustrader.exchange.bybit.constants import BybitAccountType
from nexustrader.exchange.bybit.websockets import BybitWSClient
from nexustrader.exchange.bybit.connector import (
    BybitPublicConnector,
    BybitPrivateConnector,
)
from nexustrader.exchange.bybit.exchange import BybitExchangeManager
from nexustrader.exchange.bybit.rest_api import BybitApiClient
from nexustrader.exchange.bybit.ems import BybitExecutionManagementSystem
from nexustrader.exchange.bybit.oms import BybitOrderManagementSystem
from nexustrader.exchange.bybit.factory import BybitFactory

# Auto-register factory on import
try:
    from nexustrader.exchange.registry import register_factory

    register_factory(BybitFactory())
except ImportError:
    # Registry not available yet during bootstrap
    pass

__all__ = [
    "BybitAccountType",
    "BybitWSClient",
    "BybitPublicConnector",
    "BybitExchangeManager",
    "BybitApiClient",
    "BybitPrivateConnector",
    "BybitExecutionManagementSystem",
    "BybitOrderManagementSystem",
    "BybitFactory",
]



import msgspec
from decimal import Decimal
from typing import Optional
from typing import Any, List
from nexustrader.schema import BaseMarket, Balance
from nexustrader.exchange.bitget.constants import (
    BitgetInstType,
    BitgetUtaInstType,
    BitgetOrderSide,
    BitgetOrderStatus,
    BitgetTimeInForce,
    BitgetPositionSide,
    BitgetOrderType,
)


class BitgetWsUtaArgMsg(msgspec.Struct):
    instType: BitgetUtaInstType
    topic: str
    symbol: str | None = None
    interval: str | None = None

    @property
    def message(self) -> str:
        if not self.symbol:
            return f"{self.instType.value}.{self.topic}"
        if self.interval:
            return f"{self.instType.value}.{self.topic}{self.interval}.{self.symbol}"
        return f"{self.instType.value}.{self.topic}.{self.symbol}"


class BitgetWsUtaGeneralMsg(msgspec.Struct, kw_only=True):
    event: str | None = None
    arg: BitgetWsUtaArgMsg | None = None
    code: int | None = None
    msg: str | None = None

    @property
    def is_event_data(self) -> bool:
        return self.event is not None


class BitgetGeneralResponse(msgspec.Struct, kw_only=True):
    code: str
    msg: str


class BitgetWsArgMsg(msgspec.Struct):
    instType: BitgetInstType
    channel: str
    instId: str | None = None
    coin: str | None = None

    @property
    def message(self) -> str:
        if self.instId:
            return f"{self.instType.value}.{self.channel}.{self.instId}"
        if self.coin:
            return f"{self.instType.value}.{self.channel}.{self.coin}"
        return f"{self.instType.value}.{self.channel}"


class BitgetWsGeneralMsg(msgspec.Struct, kw_only=True):
    event: str | None = None
    arg: BitgetWsArgMsg | None = None
    code: int | None = None
    msg: str | None = None

    @property
    def is_event_data(self) -> bool:
        return self.event is not None


class BookData(msgspec.Struct, array_like=True):
    px: str
    sz: str


class BitgetBooks1WsMsgData(msgspec.Struct):
    a: List[BookData]
    b: List[BookData]
    ts: str


class BitgetBooks1WsMsg(msgspec.Struct):
    data: list[BitgetBooks1WsMsgData]


class BitgetTradeWsMsgData(msgspec.Struct):
    p: str  # fill price
    S: BitgetOrderSide  # fill side
    T: str  # ts
    v: str  # fill size
    i: str  # trade id


class BitgetWsTradeWsMsg(msgspec.Struct):
    data: list[BitgetTradeWsMsgData]


class BitgetWsCandleWsMsgData(msgspec.Struct):
    start: str
    open: str
    close: str
    high: str
    low: str
    volume: str
    turnover: str


class BitgetWsCandleWsMsg(msgspec.Struct):
    data: list[BitgetWsCandleWsMsgData]


# # --- Kline ---
# class BitgetKline(msgspec.Struct):
#     openTime: int
#     open: str
#     high: str
#     low: str
#     close: str
#     volume: str
#     quoteVolume: str
#     closeTime: int

# class BitgetKlineMsg(msgspec.Struct):
#     event: str
#     arg: Dict[str, Any]
#     data: List[BitgetKline]


# # --- Trade ---
# class BitgetTrade(msgspec.Struct):
#     tradeId: str
#     price: str
#     size: str
#     side: str
#     timestamp: int

# class BitgetTradeMsg(msgspec.Struct):
#     event: str
#     arg: Dict[str, Any]
#     data: List[BitgetTrade]

# # --- Order Book ---
# class BitgetOrderBookSnapshot(msgspec.Struct):
#     asks: List[List[str]]
#     bids: List[List[str]]
#     ts: int
#     checksum: int

# class BitgetOrderBookUpdate(msgspec.Struct):
#     asks: List[List[str]]
#     bids: List[List[str]]
#     ts: int
#     checksum: int

# class BitgetOrderBookMsg(msgspec.Struct):
#     event: str
#     arg: Dict[str, Any]
#     action: str  # snapshot or update
#     data: List[BitgetOrderBookSnapshot | BitgetOrderBookUpdate]

# class BitgetOrderBook(msgspec.Struct):
#     bids: Dict[float, float] = {}
#     asks: Dict[float, float] = {}

#     def parse_orderbook(self, msg: BitgetOrderBookMsg, levels: int = 1):
#         for entry in msg.data:
#             if msg.action == "snapshot":
#                 self._handle_snapshot(entry)
#             elif msg.action == "update":
#                 self._handle_delta(entry)
#         return self._get_orderbook(levels)

#     def _handle_snapshot(self, data: BitgetOrderBookSnapshot):
#         self.bids.clear()
#         self.asks.clear()
#         for price, size in data.bids:
#             self.bids[float(price)] = float(size)
#         for price, size in data.asks:
#             self.asks[float(price)] = float(size)

#     def _handle_delta(self, data: BitgetOrderBookUpdate):
#         for price, size in data.bids:
#             price_f = float(price)
#             size_f = float(size)
#             if size_f == 0:
#                 self.bids.pop(price_f, None)
#             else:
#                 self.bids[price_f] = size_f
#         for price, size in data.asks:
#             price_f = float(price)
#             size_f = float(size)
#             if size_f == 0:
#                 self.asks.pop(price_f, None)
#             else:
#                 self.asks[price_f] = size_f

#     def _get_orderbook(self, levels: int):
#         bids = sorted(self.bids.items(), reverse=True)[:levels]
#         asks = sorted(self.asks.items())[:levels]
#         return {
#             "bids": [BookOrderData(price=price, size=size) for price, size in bids],
#             "asks": [BookOrderData(price=price, size=size) for price, size in asks],
#         }

# # --- Ticker ---
# class BitgetTicker(msgspec.Struct):
#     symbol: str
#     markPrice: str | None = None
#     indexPrice: str | None = None
#     nextFundingTime: str | None = None
#     fundingRate: str | None = None

# class BitgetTickerMsg(msgspec.Struct):
#     event: str
#     arg: Dict[str, Any]
#     data: List[Dict[str, Any]]

# # --- Balance ---
# class BitgetBalanceCoin(msgspec.Struct):
#     coin: str
#     available: str
#     frozen: str

#     def parse_to_balance(self) -> Balance:
#         locked = Decimal(self.frozen)
#         free = Decimal(self.available)
#         return Balance(asset=self.coin, locked=locked, free=free)

# class BitgetBalanceResponse(msgspec.Struct):
#     code: str
#     msg: str
#     data: List[BitgetBalanceCoin]

#     def parse_to_balances(self) -> List[Balance]:
#         return [coin.parse_to_balance() for coin in self.data]


# # --- Market Info ---
class BitgetMarketInfo(msgspec.Struct, kw_only=True, omit_defaults=True):
    # Common required fields
    symbol: str
    baseCoin: str
    quoteCoin: str
    makerFeeRate: str
    takerFeeRate: str
    minTradeUSDT: str

    # Spot-only optional fields
    minTradeAmount: Optional[str] = None
    maxTradeAmount: Optional[str] = None
    pricePrecision: Optional[str] = None
    quantityPrecision: Optional[str] = None
    quotePrecision: Optional[str] = None
    status: Optional[str] = None
    buyLimitPriceRatio: Optional[str] = None
    sellLimitPriceRatio: Optional[str] = None
    areaSymbol: Optional[str] = None
    orderQuantity: Optional[str] = None
    openTime: Optional[str] = None
    offTime: Optional[str] = None

    # Futures-only optional fields
    feeRateUpRatio: Optional[str] = None
    openCostUpRatio: Optional[str] = None
    supportMarginCoins: Optional[List[str]] = None
    minTradeNum: Optional[str] = None
    priceEndStep: Optional[str] = None
    volumePlace: Optional[str] = None
    pricePlace: Optional[str] = None
    sizeMultiplier: Optional[str] = None
    symbolType: Optional[str] = None
    maxSymbolOrderNum: Optional[str] = None
    maxProductOrderNum: Optional[str] = None
    maxPositionNum: Optional[str] = None
    symbolStatus: Optional[str] = None
    limitOpenTime: Optional[str] = None
    deliveryTime: Optional[str] = None
    deliveryStartTime: Optional[str] = None
    launchTime: Optional[str] = None
    fundInterval: Optional[str] = None
    minLever: Optional[str] = None
    maxLever: Optional[str] = None
    posLimit: Optional[str] = None
    maintainTime: Optional[str] = None
    maxMarketOrderQty: Optional[str] = None
    maxOrderQty: Optional[str] = None


class BitgetMarket(BaseMarket):
    info: BitgetMarketInfo


class BitgetOrderCancelData(msgspec.Struct):
    orderId: str
    clientOid: str


class BitgetOrderCancelResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: BitgetOrderCancelData


class BitgetOrderPlaceData(msgspec.Struct, kw_only=True):
    orderId: str
    clientOid: str | None = None


class BitgetOrderPlaceResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: BitgetOrderPlaceData


class BitgetPositionItem(msgspec.Struct):
    symbol: str
    marginCoin: str
    holdSide: BitgetPositionSide
    openDelegateSize: str
    marginSize: str
    available: str
    locked: str
    total: str
    leverage: str
    openPriceAvg: str
    marginMode: str
    posMode: str
    unrealizedPL: str
    liquidationPrice: str
    markPrice: str
    breakEvenPrice: str
    achievedProfits: str | None = None
    keepMarginRate: str | None = None
    totalFee: str | None = None
    deductedFee: str | None = None
    marginRatio: str | None = None
    assetMode: str | None = None
    uTime: str | None = None
    autoMargin: str | None = None
    cTime: str | None = None


class BitgetPositionListResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: list[BitgetPositionItem]


class BitgetOrder(msgspec.Struct, kw_only=True):
    orderId: str
    clientOid: Optional[str]
    symbol: str
    baseCoin: str
    quoteCoin: str
    size: Decimal
    price: Decimal
    state: str
    orderType: str
    side: str
    timeInForceValue: Optional[str] = None
    force: Optional[str] = None
    priceAvg: Optional[Decimal] = None
    fillPrice: Optional[Decimal] = None
    filledQty: Optional[Decimal] = None
    fee: Optional[Decimal] = None
    orderSource: Optional[str] = None
    cTime: Optional[int] = None
    uTime: Optional[int] = None
    status: Optional[str] = None


class BitgetOpenOrdersResponse(msgspec.Struct, kw_only=True):
    code: str
    msg: Optional[str]
    requestTime: int
    data: List[BitgetOrder]


class BitgetOrderHistoryItem(msgspec.Struct, kw_only=True, omit_defaults=True):
    orderId: str
    symbol: str
    price: str
    size: str
    orderType: str
    side: str
    status: str
    createTime: int
    baseCoin: Optional[str] = None
    quoteCoin: Optional[str] = None
    clientOid: Optional[str] = None
    priceAvg: Optional[str] = None
    filledAmount: Optional[str] = None
    enterPointSource: Optional[str] = None
    tradeSide: Optional[str] = None
    forceClose: Optional[bool] = None
    marginMode: Optional[str] = None
    reduceOnly: Optional[bool] = None
    presetStopSurplusPrice: Optional[str] = None
    presetStopLossPrice: Optional[str] = None
    feeDetail: Optional[str] = None
    tradeId: Optional[str] = None


class BitgetOrderHistoryResponse(msgspec.Struct, kw_only=True, omit_defaults=True):
    code: str
    msg: str
    requestTime: int
    data: List[BitgetOrderHistoryItem]


class BitgetAccountAssetItem(msgspec.Struct):
    coin: str
    available: str
    frozen: str
    locked: str
    limitAvailable: str
    uTime: str


class BitgetAccountAssetResponse(msgspec.Struct):
    code: str
    message: str
    requestTime: int
    data: List[BitgetAccountAssetItem]


class BitgetOrderModifyResponse(msgspec.Struct, kw_only=True):
    orderId: str
    clientOid: str


class BitgetResponse(msgspec.Struct, kw_only=True):
    code: str
    msg: str
    data: BitgetOrderModifyResponse
    requestTime: int


class BitgetBaseResponse(msgspec.Struct, kw_only=True):
    code: str
    msg: str
    requestTime: int
    data: Any


class BitgetKlineItem(msgspec.Struct):
    timestamp: str  # index[0]
    open: str  # index[1]
    high: str  # index[2]
    low: str  # index[3]
    close: str  # index[4]
    volume_base: str  # index[5]
    volume_quote: str  # index[6]


class BitgetKlineResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: List[List[str]]


class BitgetIndexPriceKlineItem(msgspec.Struct):
    timestamp: str
    open_price: str
    high_price: str
    low_price: str
    close_price: str
    base_volume: str
    quote_volume: str


class BitgetIndexPriceKlineResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: List[BitgetIndexPriceKlineItem]


class BitgetOrderFeeDetail(msgspec.Struct):
    feeCoin: str
    fee: str


class BitgetOrderData(msgspec.Struct, kw_only=True):
    # Common required fields (present in both spot and futures)
    instId: str
    orderId: str
    clientOid: str
    size: str
    orderType: BitgetOrderType
    force: BitgetTimeInForce
    side: BitgetOrderSide

    tradeId: Optional[str] = None
    fillTime: Optional[str] = None
    fillFee: Optional[str] = None
    fillFeeCoin: Optional[str] = None
    tradeScope: Optional[str] = None
    priceAvg: Optional[str] = None
    status: BitgetOrderStatus
    cTime: str
    uTime: str
    stpMode: str
    feeDetail: List[BitgetOrderFeeDetail]
    enterPointSource: str

    # Optional fields (may be present in spot, futures, or both)
    fillPrice: Optional[str] = None
    newSize: Optional[str] = None
    notional: Optional[str] = None
    baseVolume: Optional[str] = None
    accBaseVolume: Optional[str] = None

    # Futures-specific optional fields
    fillNotionalUsd: Optional[str] = None
    leverage: Optional[str] = None
    marginCoin: Optional[str] = None
    marginMode: Optional[str] = None
    notionalUsd: Optional[str] = None
    pnl: Optional[str] = None
    posMode: Optional[str] = None
    posSide: Optional[BitgetPositionSide] = None
    price: Optional[str] = None
    reduceOnly: Optional[str] = None
    tradeSide: Optional[str] = None
    presetStopSurplusPrice: Optional[str] = None
    totalProfits: Optional[str] = None
    presetStopLossPrice: Optional[str] = None
    cancelReason: Optional[str] = None


class BitgetOrderWsMsg(msgspec.Struct):
    action: str
    arg: BitgetWsArgMsg
    data: List[BitgetOrderData]
    ts: int


class BitgetPositionData(msgspec.Struct):
    posId: str
    instId: str
    marginCoin: str
    marginSize: str
    marginMode: str
    holdSide: BitgetPositionSide
    posMode: str
    total: str
    available: str
    frozen: str
    openPriceAvg: str
    leverage: int
    achievedProfits: str
    unrealizedPL: str
    unrealizedPLR: str
    liquidationPrice: str
    keepMarginRate: str
    marginRate: str
    cTime: str
    breakEvenPrice: str
    totalFee: str
    deductedFee: str
    markPrice: str
    uTime: str
    autoMargin: str


class BitgetPositionWsMsg(msgspec.Struct):
    action: str
    arg: BitgetWsArgMsg
    data: List[BitgetPositionData]
    ts: int


class BitgetSpotAccountData(msgspec.Struct):
    coin: str
    available: str
    frozen: str
    locked: str
    limitAvailable: str
    uTime: str

    def parse_to_balance(self) -> Balance:
        locked = Decimal(self.frozen) + Decimal(self.locked)
        free = Decimal(self.available)
        return Balance(asset=self.coin, locked=locked, free=free)


class BitgetSpotAccountWsMsg(msgspec.Struct):
    data: List[BitgetSpotAccountData]
    ts: int

    def parse_to_balances(self) -> List[Balance]:
        return [account_data.parse_to_balance() for account_data in self.data]


class BitgetFuturesAccountData(msgspec.Struct):
    marginCoin: str
    frozen: str
    available: str
    maxOpenPosAvailable: str
    maxTransferOut: str
    equity: str
    usdtEquity: str
    crossedRiskRate: str
    unrealizedPL: str

    def parse_to_balance(self) -> Balance:
        locked = Decimal(self.frozen)
        free = Decimal(self.available)
        return Balance(asset=self.marginCoin, locked=locked, free=free)


class BitgetFuturesAccountWsMsg(msgspec.Struct):
    data: List[BitgetFuturesAccountData]
    ts: int

    def parse_to_balances(self) -> List[Balance]:
        return [account_data.parse_to_balance() for account_data in self.data]


class BitgetUtaCoinData(msgspec.Struct):
    debts: str
    balance: str
    available: str
    borrow: str
    locked: str
    equity: str
    coin: str
    usdValue: str

    def parse_to_balance(self) -> Balance:
        locked = Decimal(self.locked)
        free = Decimal(self.available)
        return Balance(asset=self.coin, locked=locked, free=free)


class BitgetUtaAccountData(msgspec.Struct):
    unrealisedPnL: str
    totalEquity: str
    positionMgnRatio: str
    mmr: str
    effEquity: str
    imr: str
    mgnRatio: str
    coin: List[BitgetUtaCoinData]

    def parse_to_balances(self) -> List[Balance]:
        return [coin_data.parse_to_balance() for coin_data in self.coin]


class BitgetUtaAccountWsMsg(msgspec.Struct):
    data: List[BitgetUtaAccountData]
    ts: int

    def parse_to_balances(self) -> List[Balance]:
        balances = []
        for account_data in self.data:
            balances.extend(account_data.parse_to_balances())
        return balances


class BitgetUtaOrderFeeDetail(msgspec.Struct):
    feeCoin: str
    fee: str


class BitgetUtaOrderData(msgspec.Struct, kw_only=True):
    category: BitgetUtaInstType
    symbol: str
    orderId: str
    clientOid: str
    price: str
    qty: str
    holdMode: str
    holdSide: str
    tradeSide: str
    orderType: BitgetOrderType
    timeInForce: BitgetTimeInForce
    side: BitgetOrderSide
    marginMode: str
    marginCoin: str
    reduceOnly: str
    cumExecQty: str
    cumExecValue: str
    avgPrice: str
    totalProfit: str
    orderStatus: BitgetOrderStatus
    cancelReason: str
    leverage: str
    feeDetail: List[BitgetUtaOrderFeeDetail] | None = None
    createdTime: str
    updatedTime: str
    stpMode: str


class BitgetUtaOrderWsMsg(msgspec.Struct):
    data: List[BitgetUtaOrderData]
    ts: int


class BitgetUtaPositionData(msgspec.Struct, kw_only=True):
    symbol: str
    leverage: str
    openFeeTotal: str
    mmr: str
    breakEvenPrice: str
    available: str
    liqPrice: str
    marginMode: str
    unrealisedPnl: str
    markPrice: str
    createdTime: str
    openPriceAvg: str | None = None
    totalFundingFee: str
    updatedTime: str
    marginCoin: str
    frozen: str
    profitRate: str
    closeFeeTotal: str
    marginSize: str
    curRealisedPnl: str
    size: str
    posSide: BitgetPositionSide
    holdMode: str


class BitgetUtaPositionWsMsg(msgspec.Struct):
    data: List[BitgetUtaPositionData]
    ts: int


class BitgetWsApiArgParamsMsg(msgspec.Struct):
    orderId: str | None = None
    clientOid: str | None = None


class BitgetWsApiArgMsg(msgspec.Struct):
    id: str
    instType: str
    channel: str
    instId: str
    params: BitgetWsApiArgParamsMsg

    @property
    def is_place_order(self):
        return self.channel == "place-order"

    @property
    def is_cancel_order(self):
        return self.channel == "cancel-order"


class BitgetWsApiGeneralMsg(msgspec.Struct):
    event: str
    code: int | str
    arg: list[BitgetWsApiArgMsg] | None = None
    msg: str | None = None

    @property
    def is_success(self):
        return int(self.code) == 0

    @property
    def is_login_msg(self):
        return self.event == "login"

    @property
    def is_arg_msg(self):
        return self.arg is not None

    @property
    def is_error_msg(self):
        return int(self.code) != 0

    @property
    def error_msg(self):
        return f"code={self.code} msg={self.msg}"


class BitgetWsApiUtaArgMsg(msgspec.Struct, kw_only=True):
    orderId: str
    clientOid: str


class BitgetWsApiUtaGeneralMsg(msgspec.Struct, kw_only=True):
    event: str
    id: str | None = None
    code: str | int
    args: list[BitgetWsApiUtaArgMsg] | None = None
    msg: str | None = None

    @property
    def oid(self):
        if not self.id:
            raise ValueError("id is None")
        return self.id[1:]

    @property
    def is_cancel_order(self):
        self.id.startswith("c")

    @property
    def is_place_order(self):
        self.id.startswith("n")

    @property
    def is_success(self):
        return int(self.code) == 0

    @property
    def is_id_msg(self):
        return self.id is not None

    @property
    def is_error_msg(self):
        return int(self.code) != 0

    @property
    def error_msg(self):
        return f"code={self.code} msg={self.msg}"


class BitgetTickerResponseData(msgspec.Struct):
    category: BitgetInstType
    symbol: str
    lastPrice: str
    volume24h: str
    turnover24h: str


class BitgetTickerResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: List[BitgetTickerResponseData]


class BitgetV3PositionData(msgspec.Struct, kw_only=True):
    category: str
    symbol: str
    marginCoin: str
    holdMode: str
    posSide: str
    marginMode: str
    positionBalance: str
    available: str
    frozen: str
    total: str
    leverage: str
    curRealisedPnl: str
    avgPrice: str
    positionStatus: str
    unrealisedPnl: str
    liquidationPrice: str
    mmr: str
    profitRate: str
    markPrice: str
    breakEvenPrice: str
    totalFunding: str
    openFeeTotal: str
    closeFeeTotal: str
    createdTime: str
    updatedTime: str


class BitgetV3PositionResponseData(msgspec.Struct):
    list: List[BitgetV3PositionData] | None = None


class BitgetV3PositionResponse(msgspec.Struct):
    code: str
    msg: str
    requestTime: int
    data: BitgetV3PositionResponseData



from enum import Enum, unique
from throttled.asyncio import Throttled, rate_limiter, RateLimiterType
from throttled import Throttled as ThrottledSync
from throttled import rate_limiter as rate_limiter_sync
from nexustrader.constants import (
    AccountType,
    OrderStatus,
    PositionSide,
    OrderSide,
    TimeInForce,
    OrderType,
    KlineInterval,
    RateLimiter,
    RateLimiterSync,
    TriggerType,
)
from nexustrader.error import KlineSupportedError


class OkxWsApiOp(Enum):
    PLACE_ORDER = "order"
    BATCH_ORDERS = "batch-orders"
    CANCEL_ORDER = "cancel-order"

    @property
    def is_place_order(self):
        return self == self.PLACE_ORDER

    @property
    def is_batch_orders(self):
        return self == self.BATCH_ORDERS

    @property
    def is_cancel_order(self):
        return self == self.CANCEL_ORDER


class OkxTriggerType(Enum):
    NONE = ""
    LAST_PRICE = "last"
    INDEX_PRICE = "index"
    MARK_PRICE = "mark"


class OkxAcctLv(Enum):
    SPOT = "1"
    FUTURES = "2"
    MULTI_CURRENCY_MARGIN = "3"
    PORTFOLIO_MARGIN = "4"

    @property
    def is_spot(self):
        return self == self.SPOT

    @property
    def is_futures(self):
        return self == self.FUTURES

    @property
    def is_multi_currency_margin(self):
        return self == self.MULTI_CURRENCY_MARGIN

    @property
    def is_portfolio_margin(self):
        return self == self.PORTFOLIO_MARGIN


class OkxPositionMode(Enum):
    ONE_WAY_MODE = "net_mode"
    LONG_SHORT_MODE = "long_short_mode"

    @property
    def is_one_way_mode(self):
        return self == self.ONE_WAY_MODE

    @property
    def is_long_short_mode(self):
        return self == self.LONG_SHORT_MODE


class OkxSavingsPurchaseRedemptSide(Enum):
    PURCHASE = "purchase"
    REDEMPT = "redempt"


class OkxKlineInterval(Enum):
    SECOND_1 = "candle1s"
    MINUTE_1 = "candle1m"
    MINUTE_3 = "candle3m"
    MINUTE_5 = "candle5m"
    MINUTE_15 = "candle15m"
    MINUTE_30 = "candle30m"
    HOUR_1 = "candle1H"
    HOUR_4 = "candle4H"
    HOUR_6 = "candle6Hutc"
    HOUR_12 = "candle12Hutc"
    DAY_1 = "candle1Dutc"
    DAY_3 = "candle3Dutc"
    WEEK_1 = "candle1Wutc"
    MONTH_1 = "candle1Mutc"


class OkxInstrumentType(Enum):
    SPOT = "SPOT"
    MARGIN = "MARGIN"
    SWAP = "SWAP"
    FUTURES = "FUTURES"
    OPTION = "OPTION"
    ANY = "ANY"


class OkxInstrumentFamily(Enum):
    FUTURES = "FUTURES"
    SWAP = "SWAP"
    OPTION = "OPTION"


class OkxAccountType(AccountType):
    LIVE = "live"
    # AWS = "aws" # deprecated
    DEMO = "demo"
    LINEAR_MOCK = "linear_mock"
    INVERSE_MOCK = "inverse_mock"
    SPOT_MOCK = "spot_mock"

    @property
    def exchange_id(self):
        return "okx"

    @property
    def is_testnet(self):
        return self == OkxAccountType.DEMO

    @property
    def stream_url(self):
        return STREAM_URLS[self]

    @property
    def is_mock(self):
        return self in (self.LINEAR_MOCK, self.INVERSE_MOCK, self.SPOT_MOCK)

    @property
    def is_linear_mock(self):
        return self == self.LINEAR_MOCK

    @property
    def is_inverse_mock(self):
        return self == self.INVERSE_MOCK

    @property
    def is_spot_mock(self):
        return self == self.SPOT_MOCK


STREAM_URLS = {
    OkxAccountType.LIVE: "wss://ws.okx.com:8443/ws",
    OkxAccountType.DEMO: "wss://wspap.okx.com:8443/ws",
}

REST_URLS = {
    OkxAccountType.LIVE: "https://www.okx.com",
    OkxAccountType.DEMO: "https://www.okx.com",
}


@unique
class OkxTdMode(Enum):
    CASH = "cash"  # 现货
    CROSS = "cross"  # 全仓
    ISOLATED = "isolated"  # 逐仓
    SPOT_ISOLATED = "spot_isolated"  # 现货逐仓

    @property
    def is_cash(self):
        return self == self.CASH

    @property
    def is_cross(self):
        return self == self.CROSS

    @property
    def is_isolated(self):
        return self == self.ISOLATED

    @property
    def is_spot_isolated(self):
        return self == self.SPOT_ISOLATED


@unique
class OkxPositionSide(Enum):
    LONG = "long"
    SHORT = "short"
    NET = "net"
    NONE = ""

    def parse_to_position_side(self) -> PositionSide:
        if self == self.NET:
            return PositionSide.FLAT
        elif self == self.LONG:
            return PositionSide.LONG
        elif self == self.SHORT:
            return PositionSide.SHORT
        raise RuntimeError(f"Invalid position side: {self}")


@unique
class OkxOrderSide(Enum):
    BUY = "buy"
    SELL = "sell"

    @property
    def is_buy(self):
        return self == self.BUY

    @property
    def is_sell(self):
        return self == self.SELL


class OkxTimeInForce(Enum):
    IOC = "ioc"
    GTC = "gtc"
    FOK = "fok"


@unique
class OkxOrderType(Enum):
    MARKET = "market"
    LIMIT = "limit"
    POST_ONLY = "post_only"  # limit only, requires "px" to be provided
    FOK = "fok"  # market order if "px" is not provided, otherwise limit order
    IOC = "ioc"  # market order if "px" is not provided, otherwise limit order
    OPTIMAL_LIMIT_IOC = (
        "optimal_limit_ioc"  # Market order with immediate-or-cancel order
    )
    MMP = "mmp"  # Market Maker Protection (only applicable to Option in Portfolio Margin mode)
    MMP_AND_POST_ONLY = "mmp_and_post_only"  # Market Maker Protection and Post-only order(only applicable to Option in Portfolio Margin mode)


@unique
class OkxOrderStatus(Enum):  # "state"
    CANCELED = "canceled"
    LIVE = "live"
    PARTIALLY_FILLED = "partially_filled"
    FILLED = "filled"
    MMP_CANCELED = "mmp_canceled"


class OkxEnumParser:
    _okx_trigger_type_map = {
        OkxTriggerType.LAST_PRICE: TriggerType.LAST_PRICE,
        OkxTriggerType.INDEX_PRICE: TriggerType.INDEX_PRICE,
        OkxTriggerType.MARK_PRICE: TriggerType.MARK_PRICE,
    }

    _okx_kline_interval_map = {
        OkxKlineInterval.SECOND_1: KlineInterval.SECOND_1,
        OkxKlineInterval.MINUTE_1: KlineInterval.MINUTE_1,
        OkxKlineInterval.MINUTE_3: KlineInterval.MINUTE_3,
        OkxKlineInterval.MINUTE_5: KlineInterval.MINUTE_5,
        OkxKlineInterval.MINUTE_15: KlineInterval.MINUTE_15,
        OkxKlineInterval.MINUTE_30: KlineInterval.MINUTE_30,
        OkxKlineInterval.HOUR_1: KlineInterval.HOUR_1,
        OkxKlineInterval.HOUR_4: KlineInterval.HOUR_4,
        OkxKlineInterval.HOUR_6: KlineInterval.HOUR_6,
        OkxKlineInterval.HOUR_12: KlineInterval.HOUR_12,
        OkxKlineInterval.DAY_1: KlineInterval.DAY_1,
        OkxKlineInterval.DAY_3: KlineInterval.DAY_3,
        OkxKlineInterval.WEEK_1: KlineInterval.WEEK_1,
        OkxKlineInterval.MONTH_1: KlineInterval.MONTH_1,
    }

    _okx_order_status_map = {
        OkxOrderStatus.LIVE: OrderStatus.ACCEPTED,
        OkxOrderStatus.PARTIALLY_FILLED: OrderStatus.PARTIALLY_FILLED,
        OkxOrderStatus.FILLED: OrderStatus.FILLED,
        OkxOrderStatus.CANCELED: OrderStatus.CANCELED,
    }

    _okx_position_side_map = {
        OkxPositionSide.NET: PositionSide.FLAT,
        OkxPositionSide.LONG: PositionSide.LONG,
        OkxPositionSide.SHORT: PositionSide.SHORT,
        OkxPositionSide.NONE: None,
    }

    _okx_order_side_map = {
        OkxOrderSide.BUY: OrderSide.BUY,
        OkxOrderSide.SELL: OrderSide.SELL,
    }

    # Add reverse mapping dictionaries
    _order_status_to_okx_map = {v: k for k, v in _okx_order_status_map.items()}
    _position_side_to_okx_map = {
        PositionSide.FLAT: OkxPositionSide.NET,
        PositionSide.LONG: OkxPositionSide.LONG,
        PositionSide.SHORT: OkxPositionSide.SHORT,
    }
    _order_side_to_okx_map = {v: k for k, v in _okx_order_side_map.items()}

    _kline_interval_to_okx_map = {v: k for k, v in _okx_kline_interval_map.items()}
    _trigger_type_to_okx_map = {v: k for k, v in _okx_trigger_type_map.items()}

    @classmethod
    def parse_trigger_type(cls, trigger_type: OkxTriggerType) -> TriggerType:
        return cls._okx_trigger_type_map[trigger_type]

    @classmethod
    def parse_kline_interval(cls, interval: OkxKlineInterval) -> KlineInterval:
        return cls._okx_kline_interval_map[interval]

    # Add reverse parsing methods
    @classmethod
    def parse_order_status(cls, status: OkxOrderStatus) -> OrderStatus:
        return cls._okx_order_status_map[status]

    @classmethod
    def parse_position_side(cls, side: OkxPositionSide) -> PositionSide:
        return cls._okx_position_side_map[side]

    @classmethod
    def parse_order_side(cls, side: OkxOrderSide) -> OrderSide:
        return cls._okx_order_side_map[side]

    @classmethod
    def parse_order_type(cls, ordType: OkxOrderType) -> OrderType:
        # TODO add parameters in future to enable parsing of all other nautilus OrderType's
        match ordType:
            case OkxOrderType.MARKET:
                return OrderType.MARKET
            case OkxOrderType.LIMIT:
                return OrderType.LIMIT
            case OkxOrderType.IOC:
                return OrderType.LIMIT
            case OkxOrderType.FOK:
                return OrderType.LIMIT
            case OkxOrderType.POST_ONLY:
                return OrderType.POST_ONLY
            case _:
                raise NotImplementedError(
                    f"Cannot parse OrderType from OKX order type {ordType}"
                )

    @classmethod
    def parse_time_in_force(cls, ordType: OkxOrderType) -> TimeInForce:
        match ordType:
            case OkxOrderType.MARKET:
                return TimeInForce.GTC
            case OkxOrderType.LIMIT:
                return TimeInForce.GTC
            case OkxOrderType.POST_ONLY:
                return TimeInForce.GTC
            case OkxOrderType.FOK:
                return TimeInForce.FOK
            case OkxOrderType.IOC:
                return TimeInForce.IOC
            case _:
                raise NotImplementedError(
                    f"Cannot parse TimeInForce from OKX order type {ordType}"
                )

    @classmethod
    def to_okx_order_status(cls, status: OrderStatus) -> OkxOrderStatus:
        return cls._order_status_to_okx_map[status]

    @classmethod
    def to_okx_position_side(cls, side: PositionSide) -> OkxPositionSide:
        return cls._position_side_to_okx_map[side]

    @classmethod
    def to_okx_order_side(cls, side: OrderSide) -> OkxOrderSide:
        return cls._order_side_to_okx_map[side]

    @classmethod
    def to_okx_trigger_type(cls, trigger_type: TriggerType) -> OkxTriggerType:
        return cls._trigger_type_to_okx_map[trigger_type]

    @classmethod
    def to_okx_order_type(
        cls, order_type: OrderType, time_in_force: TimeInForce
    ) -> OkxOrderType:
        if order_type == OrderType.MARKET:
            return OkxOrderType.MARKET
        elif order_type == OrderType.POST_ONLY:
            return OkxOrderType.POST_ONLY

        match time_in_force:
            case TimeInForce.GTC:
                return OkxOrderType.LIMIT  # OKX limit orders are GTC by default
            case TimeInForce.FOK:
                return OkxOrderType.FOK
            case TimeInForce.IOC:
                return OkxOrderType.IOC
            case _:
                raise RuntimeError(
                    f"Could not determine OKX order type from order_type {order_type} and time_in_force {time_in_force}, valid OKX order types are: {list(OkxOrderType)}",
                )

    @classmethod
    def to_okx_kline_interval(cls, interval: KlineInterval) -> OkxKlineInterval:
        if interval not in cls._kline_interval_to_okx_map:
            raise KlineSupportedError(
                f"Kline interval {interval} is not supported by OKX"
            )
        return cls._kline_interval_to_okx_map[interval]


class OkxRateLimiter(RateLimiter):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, Throttled] = {
            "/api/v5/account/balance": Throttled(
                quota=rate_limiter.per_sec(5),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/account/positions": Throttled(
                quota=rate_limiter.per_sec(5),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/order": Throttled(
                quota=rate_limiter.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/cancel-order": Throttled(
                quota=rate_limiter.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/candles": Throttled(
                quota=rate_limiter.per_sec(20),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/history-candles": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/amend-order": Throttled(
                quota=rate_limiter.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/account/config": Throttled(
                quota=rate_limiter.per_sec(2),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/history-index-candles": Throttled(
                quota=rate_limiter.per_sec(4),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/batch-orders": Throttled(
                quota=rate_limiter.per_sec(150),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/tickers": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/ticker": Throttled(
                quota=rate_limiter.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/ws/order": Throttled(
                quota=rate_limiter.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/ws/cancel": Throttled(
                quota=rate_limiter.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
        }

    def __call__(self, endpoint: str) -> Throttled:
        return self._throttled[endpoint]


class OkxRateLimiterSync(RateLimiterSync):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, ThrottledSync] = {
            "/api/v5/account/balance": ThrottledSync(
                quota=rate_limiter_sync.per_sec(5),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/account/positions": ThrottledSync(
                quota=rate_limiter_sync.per_sec(5),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/order": ThrottledSync(
                quota=rate_limiter_sync.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/cancel-order": ThrottledSync(
                quota=rate_limiter_sync.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/candles": ThrottledSync(
                quota=rate_limiter_sync.per_sec(20),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/history-candles": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/amend-order": ThrottledSync(
                quota=rate_limiter_sync.per_sec(30),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/account/config": ThrottledSync(
                quota=rate_limiter_sync.per_sec(2),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/history-index-candles": ThrottledSync(
                quota=rate_limiter_sync.per_sec(4),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/trade/batch-orders": ThrottledSync(
                quota=rate_limiter_sync.per_sec(150),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/tickers": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
            "/api/v5/market/ticker": ThrottledSync(
                quota=rate_limiter_sync.per_sec(10),
                timeout=60 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
        }

    def __call__(self, endpoint: str) -> ThrottledSync:
        return self._throttled[endpoint]


def strip_uuid_hyphens(uuid_str: str) -> str:
    """Remove hyphens from UUID string for OKX API compatibility."""
    return uuid_str.replace("-", "")


def restore_uuid_hyphens(uuid_str: str) -> str:
    """Restore hyphens to UUID string from OKX API response."""
    if len(uuid_str) != 32:
        return uuid_str  # Return as-is if not a valid stripped UUID
    return f"{uuid_str[:8]}-{uuid_str[8:12]}-{uuid_str[12:16]}-{uuid_str[16:20]}-{uuid_str[20:]}"



from nexustrader.base.exchange import ExchangeManager
from nexustrader.base.ws_client import WSClient
from nexustrader.base.api_client import ApiClient
from nexustrader.base.oms import OrderManagementSystem
from nexustrader.base.ems import ExecutionManagementSystem
from nexustrader.base.sms import SubscriptionManagementSystem
from nexustrader.base.connector import (
    PublicConnector,
    PrivateConnector,
    MockLinearConnector,
)
from nexustrader.base.retry import RetryManager


__all__ = [
    "ExchangeManager",
    "WSClient",
    "ApiClient",
    "OrderManagementSystem",
    "ExecutionManagementSystem",
    "PublicConnector",
    "SubscriptionManagementSystem",
    "PrivateConnector",
    "MockLinearConnector",
    "RetryManager",
]



from datetime import timedelta
from typing import Dict, TypedDict, NotRequired
from nexustrader.constants import (
    AccountType,
    OrderStatus,
    PositionSide,
    OrderSide,
    TimeInForce,
    OrderType,
    KlineInterval,
)
from enum import Enum
from nexustrader.error import KlineSupportedError
from throttled.asyncio import Throttled, rate_limiter, RateLimiterType
from throttled import Throttled as ThrottledSync
from throttled import rate_limiter as rate_limiter_sync


def oid_to_cloid_hex(oid: str) -> str:
    """Convert NexusTrader OidGen's decimal OID to 128-bit hex string.

    Produces a lowercase hex string with a `0x` prefix and exactly 32 hex
    characters (128 bits), e.g. `0x0000...abcd`.

    Notes:
    - OidGen always emits a positive decimal string (timestamp+seq+shard), so
      negativity checks are unnecessary.
    - The numeric value (currently ~20 decimal digits) is far below 2^128, so
      overflow checks are unnecessary.
    - No stripping is performed because OidGen-controlled inputs have no
      surrounding whitespace.
    """
    n = int(oid)
    return f"0x{n:032x}"


class HyperLiquidAccountType(AccountType):
    MAINNET = "mainnet"
    TESTNET = "testnet"
    LINEAR_MOCK = "linear_mock"
    INVERSE_MOCK = "inverse_mock"
    SPOT_MOCK = "spot_mock"

    @property
    def exchange_id(self):
        return "hyperliquid"

    @property
    def is_testnet(self):
        return self == self.TESTNET

    @property
    def ws_url(self):
        if self.is_testnet:
            return "wss://api.hyperliquid-testnet.xyz/ws"
        return "wss://api.hyperliquid.xyz/ws"

    @property
    def rest_url(self):
        if self.is_testnet:
            return "https://api.hyperliquid-testnet.xyz"
        return "https://api.hyperliquid.xyz"

    @property
    def is_mock(self):
        return self in (self.LINEAR_MOCK, self.INVERSE_MOCK, self.SPOT_MOCK)

    @property
    def is_linear_mock(self):
        return self == self.LINEAR_MOCK

    @property
    def is_inverse_mock(self):
        return self == self.INVERSE_MOCK

    @property
    def is_spot_mock(self):
        return self == self.SPOT_MOCK


class HyperLiquidTimeInForce(Enum):
    GTC = "Gtc"
    IOC = "Ioc"
    ALO = "Alo"  # Post Only


class HyperLiquidFillDirection(Enum):
    OPEN_LONG = "Open Long"
    OPEN_SHORT = "Open Short"
    CLOSE_LONG = "Close Long"
    CLOSE_SHORT = "Close Short"
    BUY = "Buy"
    SELL = "Sell"


class HyperLiquidOrderSide(Enum):
    BUY = "B"  # Bid
    SELL = "A"  # Ask

    @property
    def is_buy(self):
        return self == self.BUY

    @property
    def is_sell(self):
        return self == self.SELL


class HyperLiquidOrderStatusType(Enum):
    """
    https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/info-endpoint#query-order-status-by-oid-or-cloid
    """

    OPEN = "open"
    FILLED = "filled"
    CANCELED = "canceled"
    TRIGGERED = "triggered"
    REJECTED = "rejected"
    MARGIN_CANCELED = "marginCanceled"
    VAULT_WITHDRAWAL_CANCELED = "vaultWithdrawalCanceled"
    OPEN_INTEREST_CAP_CANCELED = "openInterestCapCanceled"
    SELF_TRADE_CANCELED = "selfTradeCanceled"
    REDUCE_ONLY_CANCELED = "reduceOnlyCanceled"
    SIBLING_FILLED_CANCELED = "siblingFilledCanceled"
    DELISTED_CANCELED = "delistedCanceled"
    LIQUIDATED_CANCELED = "liquidatedCanceled"
    SCHEDULED_CANCEL = "scheduledCancel"
    TICK_REJECTED = "tickRejected"
    MIN_TRADE_NTL_REJECTED = "minTradeNtlRejected"
    PERP_MARGIN_REJECTED = "perpMarginRejected"
    REDUCE_ONLY_REJECTED = "reduceOnlyRejected"
    BAD_ALO_PX_REJECTED = "badAloPxRejected"
    IOC_CANCEL_REJECTED = "iocCancelRejected"
    BAD_TRIGGER_PX_REJECTED = "badTriggerPxRejected"
    MARKET_ORDER_NO_LIQUIDITY_REJECTED = "marketOrderNoLiquidityRejected"
    POSITION_INCREASE_AT_OPEN_INTEREST_CAP_REJECTED = (
        "positionIncreaseAtOpenInterestCapRejected"
    )
    POSITION_FLIP_AT_OPEN_INTEREST_CAP_REJECTED = (
        "positionFlipAtOpenInterestCapRejected"
    )
    TOO_AGGRESSIVE_AT_OPEN_INTEREST_CAP_REJECTED = (
        "tooAggressiveAtOpenInterestCapRejected"
    )
    OPEN_INTEREST_INCREASE_REJECTED = "openInterestIncreaseRejected"
    INSUFFICIENT_SPOT_BALANCE_REJECTED = "insufficientSpotBalanceRejected"
    ORACLE_REJECTED = "oracleRejected"
    PERP_MAX_POSITION_REJECTED = "perpMaxPositionRejected"


class HyperLiquidKlineInterval(Enum):
    MINUTE_1 = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"
    HOUR_1 = "1h"
    HOUR_2 = "2h"
    HOUR_4 = "4h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"
    DAY_1 = "1d"
    WEEK_1 = "1w"
    MONTH_1 = "1M"


class HyperLiquidRateLimiter:
    """Rate limiter for Hyperliquid API"""

    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: Dict[str, Throttled] = {
            "/exchange": Throttled(
                quota=rate_limiter.per_duration(timedelta(seconds=60), limit=1200),
                timeout=600 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
        }

    def __call__(self, endpoint: str) -> Throttled:
        return self._throttled[endpoint]


class HyperLiquidRateLimiterSync:
    """Synchronous rate limiter for Hyperliquid API"""

    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: Dict[str, Throttled] = {
            "/info": ThrottledSync(
                quota=rate_limiter_sync.per_duration(timedelta(seconds=60), limit=1200),
                timeout=600 if enable_rate_limit else -1,
                using=RateLimiterType.GCRA.value,
            ),
        }

    def __call__(self, endpoint: str) -> Throttled:
        return self._throttled[endpoint]


class HyperLiquidOrderLimitTypeRequest(TypedDict):
    tif: str  # "Alo" other exchange called post only | Ioc (Immediate or Cancel) | Gtc (Good Till Cancel)


class HyperLiquidOrderTriggerTypeRequest(TypedDict):
    isMarket: bool  # True for market order, False for limit order
    triggerPx: str  # Trigger price for stop orders, empty for limit orders
    tpsl: str  # tp | sl


class HyperLiquidOrderTypeRequest(TypedDict):
    limit: NotRequired[HyperLiquidOrderLimitTypeRequest]  # Limit order type
    trigger: NotRequired[HyperLiquidOrderTriggerTypeRequest]


class HyperLiquidOrderRequest(TypedDict):
    """
    HyperLiquid order request schema


    a: int  asset
    b: bool isBuy
    p: str  price
    s: str  size
    r: bool  reduceOnly
    t: HyperLiquidOrderTypeRequest
    c: NotRequired[str]  clientOrderId
    """

    a: int  # asset
    b: bool  # isBuy
    p: str  # price
    s: str  # size
    r: bool  # reduceOnly
    t: HyperLiquidOrderTypeRequest
    c: NotRequired[str]  # clientOrderId


class HyperLiquidOrderCancelRequest(TypedDict):
    a: int  # asset
    o: int  # oid  # orderId


class HyperLiquidCloidCancelRequest(TypedDict):
    asset: int  # asset
    cloid: str  # clientOrderId


class HyperLiquidEnumParser:
    _hyperliquid_order_status_map = {
        HyperLiquidOrderStatusType.OPEN: OrderStatus.ACCEPTED,
        HyperLiquidOrderStatusType.FILLED: OrderStatus.FILLED,
        HyperLiquidOrderStatusType.CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.TRIGGERED: OrderStatus.ACCEPTED,
        HyperLiquidOrderStatusType.REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.MARGIN_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.VAULT_WITHDRAWAL_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.OPEN_INTEREST_CAP_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.SELF_TRADE_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.REDUCE_ONLY_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.SIBLING_FILLED_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.DELISTED_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.LIQUIDATED_CANCELED: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.SCHEDULED_CANCEL: OrderStatus.CANCELED,
        HyperLiquidOrderStatusType.TICK_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.MIN_TRADE_NTL_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.PERP_MARGIN_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.REDUCE_ONLY_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.BAD_ALO_PX_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.IOC_CANCEL_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.BAD_TRIGGER_PX_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.MARKET_ORDER_NO_LIQUIDITY_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.POSITION_INCREASE_AT_OPEN_INTEREST_CAP_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.POSITION_FLIP_AT_OPEN_INTEREST_CAP_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.TOO_AGGRESSIVE_AT_OPEN_INTEREST_CAP_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.OPEN_INTEREST_INCREASE_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.INSUFFICIENT_SPOT_BALANCE_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.ORACLE_REJECTED: OrderStatus.FAILED,
        HyperLiquidOrderStatusType.PERP_MAX_POSITION_REJECTED: OrderStatus.FAILED,
    }

    _hyperliquid_kline_interval_map = {
        HyperLiquidKlineInterval.MINUTE_1: KlineInterval.MINUTE_1,
        HyperLiquidKlineInterval.MINUTE_3: KlineInterval.MINUTE_3,
        HyperLiquidKlineInterval.MINUTE_5: KlineInterval.MINUTE_5,
        HyperLiquidKlineInterval.MINUTE_15: KlineInterval.MINUTE_15,
        HyperLiquidKlineInterval.MINUTE_30: KlineInterval.MINUTE_30,
        HyperLiquidKlineInterval.HOUR_1: KlineInterval.HOUR_1,
        HyperLiquidKlineInterval.HOUR_2: KlineInterval.HOUR_2,
        HyperLiquidKlineInterval.HOUR_4: KlineInterval.HOUR_4,
        HyperLiquidKlineInterval.HOUR_8: KlineInterval.HOUR_8,
        HyperLiquidKlineInterval.HOUR_12: KlineInterval.HOUR_12,
        HyperLiquidKlineInterval.DAY_1: KlineInterval.DAY_1,
        HyperLiquidKlineInterval.WEEK_1: KlineInterval.WEEK_1,
        HyperLiquidKlineInterval.MONTH_1: KlineInterval.MONTH_1,
    }

    _kline_interval_to_hyperliquid_map = {
        v: k for k, v in _hyperliquid_kline_interval_map.items()
    }

    _hyperliquid_time_in_force_map = {
        HyperLiquidTimeInForce.GTC: TimeInForce.GTC,
        HyperLiquidTimeInForce.IOC: TimeInForce.IOC,
    }

    _time_in_force_to_hyperliquid_map = {
        v: k for k, v in _hyperliquid_time_in_force_map.items()
    }

    @classmethod
    def parse_order_status(cls, status: HyperLiquidOrderStatusType) -> OrderStatus:
        """Convert HyperLiquidOrderStatusType to OrderStatus"""
        return cls._hyperliquid_order_status_map[status]

    @classmethod
    def parse_kline_interval(cls, interval: HyperLiquidKlineInterval) -> KlineInterval:
        """Convert KlineInterval to HyperLiquidKlineInterval"""
        return cls._hyperliquid_kline_interval_map[interval]

    @classmethod
    def to_hyperliquid_kline_interval(
        cls, interval: KlineInterval
    ) -> HyperLiquidKlineInterval:
        """Convert KlineInterval to HyperLiquidKlineInterval"""
        if interval not in cls._kline_interval_to_hyperliquid_map:
            raise KlineSupportedError(
                f"Unsupported kline interval: {interval}. Supported intervals: {list(cls._kline_interval_to_hyperliquid_map.keys())}"
            )
        return cls._kline_interval_to_hyperliquid_map[interval]

    @classmethod
    def parse_time_in_force(cls, time_in_force: HyperLiquidTimeInForce) -> TimeInForce:
        """Convert HyperLiquidTimeInForce to TimeInForce"""
        return cls._hyperliquid_time_in_force_map[time_in_force]

    @classmethod
    def to_hyperliquid_time_in_force(
        cls, time_in_force: TimeInForce
    ) -> HyperLiquidTimeInForce:
        """Convert TimeInForce to HyperLiquidTimeInForce"""
        return cls._time_in_force_to_hyperliquid_map[time_in_force]



from nexustrader.exchange.binance import BinanceAccountType
from nexustrader.exchange.bybit import BybitAccountType
from nexustrader.exchange.okx import OkxAccountType
from nexustrader.exchange.hyperliquid import HyperLiquidAccountType
from nexustrader.exchange.bitget import BitgetAccountType

__all__ = [
    "BinanceAccountType",
    "BybitAccountType",
    "OkxAccountType",
    "HyperLiquidAccountType",
    "BitgetAccountType",
]
